# -*- coding: utf-8 -*-

__revision__ = '$Id: scriptnames.py 7386 2006-04-30 11:12:55Z matt $'

scriptNames = {
	u'Yiii': u'Yi',
	u'Telu': u'Telugu',
	u'Taml': u'Tamil',
	u'Guru': u'Gurmukhi',
	u'Hebr': u'Hebrew',
	u'Geor': u'Georgian (Mkhedruli)',
	u'Ugar': u'Ugaritic',
	u'Cyrl': u'Cyrillic',
	u'Hrkt': u'Kanji & Kana',
	u'Armn': u'Armenian',
	u'Runr': u'Runic',
	u'Khmr': u'Khmer',
	u'Latn': u'Latin',
	u'Hani': u'Han (Hanzi, Kanji, Hanja)',
	u'Ital': u'Old Italic (Etruscan, Oscan, etc.)',
	u'Hano': u'Hanunoo (Hanunóo)',
	u'Ethi': u'Ethiopic (Ge\'ez)',
	u'Gujr': u'Gujarati',
	u'Hang': u'Hangul',
	u'Arab': u'Arabic',
	u'Thaa': u'Thaana',
	u'Buhd': u'Buhid',
	u'Sinh': u'Sinhala',
	u'Orya': u'Oriya',
	u'Hans': u'Han (Simplified variant)',
	u'Thai': u'Thai',
	u'Cprt': u'Cypriot',
	u'Linb': u'Linear B',
	u'Hant': u'Han (Traditional variant)',
	u'Osma': u'Osmanya',
	u'Mong': u'Mongolian',
	u'Deva': u'Devanagari (Nagari)',
	u'Laoo': u'Lao',
	u'Tagb': u'Tagbanwa',
	u'Hira': u'Hiragana',
	u'Bopo': u'Bopomofo',
	u'Goth': u'Gothic',
	u'Tale': u'Tai Le',
	u'Mymr': u'Myanmar (Burmese)',
	u'Tglg': u'Tagalog',
	u'Grek': u'Greek',
	u'Mlym': u'Malayalam',
	u'Cher': u'Cherokee',
	u'Tibt': u'Tibetan',
	u'Kana': u'Katakana',
	u'Syrc': u'Syriac',
	u'Cans': u'Unified Canadian Aboriginal Syllabics',
	u'Beng': u'Bengali',
	u'Limb': u'Limbu',
	u'Ogam': u'Ogham',
	u'Knda': u'Kannada',
}

# EOF



#! /usr/bin/env python
# 
# Search for a track by title (and optionally by artist name).
#
# Usage:
#	python findtrack.py 'track name' ['artist name']
#
# $Id: findtrack.py 201 2006-03-27 14:43:13Z matt $
#
import sys
import logging
from musicbrainz2.webservice import Query, TrackFilter, WebServiceError

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
	

if len(sys.argv) < 2:
	print "Usage: findtrack.py 'track name' ['artist name']"
	sys.exit(1)

if len(sys.argv) > 2:
	artistName = sys.argv[2]
else:
	artistName = None

q = Query()

try:
	f = TrackFilter(title=sys.argv[1], artistName=artistName)
	results = q.getTracks(f)
except WebServiceError, e:
	print 'Error:', e
	sys.exit(1)


for result in results:
	track = result.track
	print "Score     :", result.score
	print "Id        :", track.id
	print "Title     :", track.title
	print "Artist    :", track.artist.name
	print

# EOF



#! /usr/bin/env python
#
# Retrieve a release group by ID and display it.
#
# Usage:
#	python getreleasegroup.py releasegroup-id
#
# Interesting release group IDs for testing:
# http://musicbrainz.org/release-group/e49e2f8a-94c0-3dcf-8ce6-9bc52a1a7867
# http://musicbrainz.org/release-group/c7a0fc4d-b6a0-3a43-9e25-4052e4fe33b2
# http://musicbrainz.org/release-group/055be730-dcad-31bf-b550-45ba9c202aa3
# http://musicbrainz.org/release-group/963eac15-e3da-3a92-aa5c-2ec23bfb6ec2
#
#
import sys
import logging
import musicbrainz2.webservice as ws
import musicbrainz2.utils as u

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


if len(sys.argv) < 2:
	print "Usage: getreleasegroup.py releasegroup-id"
	sys.exit(1)

q = ws.Query()

try:
	# Additionally to the release itself, we want the server to include
	# the release's artist, all release events, associated discs and
	# the track list.
	#
	inc = ws.ReleaseGroupIncludes(artist=True, releases=True, tags=False)
	releaseGroup = q.getReleaseGroupById(sys.argv[1], inc)
except ws.WebServiceError, e:
	print 'Error:', e
	sys.exit(1)


print "Id          :", releaseGroup.id
print "Title       :", releaseGroup.title
print "Type        :", releaseGroup.type


# Print the main artist of this release group.
#
if releaseGroup.artist:
	print
	print "Artist:"
	print "  Id        :", releaseGroup.artist.id
	print "  Name      :", releaseGroup.artist.name
	print "  SortName  :", releaseGroup.artist.sortName


# Show the releases contained by this release group.
#
for release in releaseGroup.releases:
	print
	print "Release:"
	print "  Id        :", release.id
	print "  Title     :", release.title
	print "  Types     :", release.types

# EOF



"""The MusicBrainz domain model.

These classes are part of the MusicBrainz domain model. They may be used
by other modules and don't contain any network or other I/O code. If you
want to request data from the web service, please have a look at
L{musicbrainz2.webservice}.

The most important classes, usually acting as entry points, are
L{Artist}, L{Release}, and L{Track}.

@var VARIOUS_ARTISTS_ID: The ID of the special 'Various Artists' artist.

@var NS_MMD_1: Default namespace prefix for all MusicBrainz metadata.
@var NS_REL_1: Namespace prefix for relations.
@var NS_EXT_1: Namespace prefix for MusicBrainz extensions.

@see: L{musicbrainz2.webservice}

@author: Matthias Friedrich <matt@mafr.de>
"""
try:
	set
except NameError:
	from sets import Set as set

__revision__ = '$Id: model.py 12829 2010-09-15 12:00:11Z luks $'

__all__ = [
	'VARIOUS_ARTISTS_ID', 'NS_MMD_1', 'NS_REL_1', 'NS_EXT_1', 
	'Entity', 'Artist', 'Release', 'Track', 'User', 'ReleaseGroup',
	'Relation', 'Disc', 'ReleaseEvent', 'Label', 'Tag', 'Rating',
	'AbstractAlias', 'ArtistAlias', 'LabelAlias',
]


VARIOUS_ARTISTS_ID = 'http://musicbrainz.org/artist/89ad4ac3-39f7-470e-963a-56509c546377'

# Namespace URI prefixes
#
NS_MMD_1 = 'http://musicbrainz.org/ns/mmd-1.0#'
NS_REL_1 = 'http://musicbrainz.org/ns/rel-1.0#'
NS_EXT_1 = 'http://musicbrainz.org/ns/ext-1.0#'


class Entity(object):
	"""A first-level MusicBrainz class.

	All entities in MusicBrainz have unique IDs (which are absolute URIs)
	as well as any number of L{relations <Relation>} to other entities
	and free text tags. This class is abstract and should not be
	instantiated.

	Relations are differentiated by their I{target type}, that means,
	where they link to. MusicBrainz currently supports four target types
	(artists, releases, tracks, and URLs) each identified using a URI.
	To get all relations with a specific target type, you can use
	L{getRelations} and pass one of the following constants as the
	parameter:

	 - L{Relation.TO_ARTIST}
	 - L{Relation.TO_RELEASE}
	 - L{Relation.TO_TRACK}
	 - L{Relation.TO_URL}

	@see: L{Relation}
	"""

	def __init__(self, id_=None):
		"""Constructor.

		This should only used by derived classes.

		@param id_: a string containing an absolute URI
		"""
		self._id = id_
		self._relations = { }
		self._tags = { }
		self._rating = Rating()

	def getId(self):
		"""Returns a MusicBrainz ID.

		@return: a string containing a URI, or None
		"""
		return self._id

	def setId(self, value):
		"""Sets a MusicBrainz ID.

		@param value: a string containing an absolute URI 
		"""
		self._id = value

	id = property(getId, setId, doc='The MusicBrainz ID.')

	def getRelations(self, targetType=None, relationType=None,
			requiredAttributes=(), direction=None):
		"""Returns a list of relations.

		If C{targetType} is given, only relations of that target
		type are returned. For MusicBrainz, the following target
		types are defined:
		 - L{Relation.TO_ARTIST}
		 - L{Relation.TO_RELEASE}
		 - L{Relation.TO_TRACK}
		 - L{Relation.TO_URL}

		If C{targetType} is L{Relation.TO_ARTIST}, for example,
		this method returns all relations between this Entity and
		artists.

		You may use the C{relationType} parameter to further restrict
		the selection. If it is set, only relations with the given
		relation type are returned. The C{requiredAttributes} sequence
		lists attributes that have to be part of all returned relations.

		If C{direction} is set, only relations with the given reading
		direction are returned. You can use the L{Relation.DIR_FORWARD},
		L{Relation.DIR_BACKWARD}, and L{Relation.DIR_NONE} constants
		for this.

		@param targetType: a string containing an absolute URI, or None
		@param relationType: a string containing an absolute URI, or None
		@param requiredAttributes: a sequence containing absolute URIs
		@param direction: one of L{Relation}'s direction constants
		@return: a list of L{Relation} objects

		@see: L{Entity}
		"""
		allRels = [ ]
		if targetType is not None:
			allRels = self._relations.setdefault(targetType, [ ])
		else:
			for (k, relList) in self._relations.items():
				for rel in relList:
					allRels.append(rel)

		# Filter for direction.
		#
		if direction is not None:
			allRels = [r for r in allRels if r.getDirection() == direction]

		# Filter for relation type.
		#
		if relationType is None:
			return allRels
		else:
			allRels = [r for r in allRels if r.getType() == relationType]

		# Now filer for attribute type.
		#
		tmp = []
		required = set(iter(requiredAttributes))

		for r in allRels:
			attrs = set(iter(r.getAttributes()))
			if required.issubset(attrs):
				tmp.append(r)
		return tmp


	def getRelationTargets(self, targetType=None, relationType=None,
			requiredAttributes=(), direction=None):
		"""Returns a list of relation targets.

		The arguments work exactly like in L{getRelations}, but
		instead of L{Relation} objects, the matching relation
		targets are returned. This can be L{Artist}, L{Release},
		or L{Track} objects, depending on the relations.

		As a special case, URL strings are returned if the target
		is an URL.

		@param targetType: a string containing an absolute URI, or None
		@param relationType: a string containing an absolute URI, or None
		@param requiredAttributes: a sequence containing absolute URIs
		@param direction: one of L{Relation}'s direction constants
		@return: a list of objects, depending on the relation

		@see: L{getRelations}
		"""
		ret = [ ]
		rels =  self.getRelations(targetType, relationType,
			requiredAttributes, direction)

		for r in rels:
			if r.getTargetType() == Relation.TO_URL:
				ret.append(r.getTargetId())
			else:
				ret.append(r.getTarget())

		return ret


	def addRelation(self, relation):
		"""Adds a relation.

		This method adds C{relation} to the list of relations. The
		given relation has to be initialized, at least the target
		type has to be set.

		@param relation: the L{Relation} object to add

		@see: L{Entity}
		"""
		assert relation.getType is not None
		assert relation.getTargetType is not None
		assert relation.getTargetId is not None
		l = self._relations.setdefault(relation.getTargetType(), [ ])
		l.append(relation)


	def getRelationTargetTypes(self):
		"""Returns a list of target types available for this entity.

		Use this to find out to which types of targets this entity
		has relations. If the entity only has relations to tracks and
		artists, for example, then a list containg the strings
		L{Relation.TO_TRACK} and L{Relation.TO_ARTIST} is returned.

		@return: a list of strings containing URIs

		@see: L{getRelations}
		"""
		return self._relations.keys()

	def getTag(self, value):
		"""Return the tag with the given value (aka the tag's name).

		@return: the L{Tag} with the given name or raises a KeyError
		"""
		return self._tags[value]

	def getTags(self):
		"""Return all tags attached to this Entity.

		@return: a list of L{Tag} objects
		"""
		return self._tags.values()

	tags = property(getTags, doc='The tags for this entity.')

	def addTag(self, tag):
		"""Add a new tag.

		This merges an existing tag with the same name.

		@param tag: the L{Tag} object to add

		@see: L{getTags}
		"""
		if self._tags.has_key(tag.value):
			existing = self._tags[tag.value]
			existing.count += tag.count
		else:
			self._tags[tag.value] = tag

	def getRating(self):
		"""Return the rating of this Entity.
		0 = Unrated
		1 - 5 = Rating

		@return: rating
		"""
		return self._rating

	rating = property(getRating, doc='The rating for this entity.')

	def setRating(self, value):
		self._rating = value
		

class Artist(Entity):
	"""Represents an artist.

	Artists in MusicBrainz can have a type. Currently, this type can
	be either Person or Group for which the following URIs are assigned:

	 - C{http://musicbrainz.org/ns/mmd-1.0#Person}
	 - C{http://musicbrainz.org/ns/mmd-1.0#Group}

	Use the L{TYPE_PERSON} and L{TYPE_GROUP} constants for comparison.
	"""
	TYPE_PERSON = NS_MMD_1 + 'Person'
	TYPE_GROUP = NS_MMD_1 + 'Group'

	def __init__(self, id_=None, type_=None, name=None, sortName=None):
		"""Constructor.

		@param id_: a string containing an absolute URI
		@param type_: a string containing an absolute URI
		@param name: a string containing the artist's name
		@param sortName: a string containing the artist's sort name
		"""
		Entity.__init__(self, id_)
		self._type = type_
		self._name = name
		self._sortName = sortName
		self._disambiguation = None
		self._beginDate = None
		self._endDate = None
		self._aliases = [ ]
		self._releases = [ ]
		self._releasesCount = None
		self._releasesOffset = None
		self._releaseGroups = [ ]
		self._releaseGroupsCount = None
		self._releaseGroupsOffset = None

	def getType(self):
		"""Returns the artist's type.

		@return: a string containing an absolute URI, or None 
		"""
		return self._type

	def setType(self, type_):
		"""Sets the artist's type.

		@param type_: a string containing an absolute URI
		"""
		self._type = type_

	type = property(getType, setType, doc="The artist's type.")

	def getName(self):
		"""Returns the artist's name.

		@return: a string containing the artist's name, or None
		"""
		return self._name

	def setName(self, name):
		"""Sets the artist's name.

		@param name: a string containing the artist's name
		"""
		self._name = name

	name = property(getName, setName, doc="The artist's name.")

	def getSortName(self):
		"""Returns the artist's sort name.

		The sort name is the artist's name in a special format which
		is better suited for lexicographic sorting. The MusicBrainz
		style guide specifies this format.

		@see: U{The MusicBrainz Style Guidelines
			<http://musicbrainz.org/style.html>}
		"""
		return self._sortName

	def setSortName(self, sortName):
		"""Sets the artist's sort name.

		@param sortName: a string containing the artist's sort name

		@see: L{getSortName}
		"""
		self._sortName = sortName

	sortName = property(getSortName, setSortName,
		doc="The artist's sort name.")

	def getDisambiguation(self):
		"""Returns the disambiguation attribute.

		This attribute may be used if there is more than one artist
		with the same name. In this case, disambiguation attributes
		are added to the artists' names to keep them apart.

		For example, there are at least three bands named 'Vixen'.
		Each band has a different disambiguation in the MusicBrainz
		database, like 'Hip-hop' or 'all-female rock/glam band'.

		@return: a disambiguation string, or None

		@see: L{getUniqueName}
		"""
		return self._disambiguation

	def setDisambiguation(self, disambiguation):
		"""Sets the disambiguation attribute.

		@param disambiguation: a disambiguation string

		@see: L{getDisambiguation}, L{getUniqueName}
		"""
		self._disambiguation = disambiguation

	disambiguation = property(getDisambiguation, setDisambiguation,
		doc="The disambiguation comment.")

	def getUniqueName(self):
		"""Returns a unique artist name (using disambiguation).

		This method returns the artist name together with the
		disambiguation attribute in parenthesis if it exists.
		Example: 'Vixen (Hip-hop)'.

		@return: a string containing the unique name

		@see: L{getDisambiguation}
		"""
		d = self.getDisambiguation() 
		if d is not None and d.strip() != '':
			return '%s (%s)' % (self.getName(), d)
		else: 
			return self.getName()

	def getBeginDate(self):
		"""Returns the birth/foundation date.

		The definition of the I{begin date} depends on the artist's
		type. For persons, this is the day of birth, for groups it
		is the day the group was founded.

		The returned date has the format 'YYYY', 'YYYY-MM', or 
		'YYYY-MM-DD', depending on how much detail is known.

		@return: a string containing the date, or None

		@see: L{getType}
		"""
		return self._beginDate

	def setBeginDate(self, dateStr):
		"""Sets the begin/foundation date.

		@param dateStr: a date string

		@see: L{getBeginDate}
		"""
		self._beginDate = dateStr

	beginDate = property(getBeginDate, setBeginDate,
		doc="The begin/foundation date.")

	def getEndDate(self):
		"""Returns the death/dissolving date.

		The definition of the I{end date} depends on the artist's
		type. For persons, this is the day of death, for groups it
		is the day the group was dissolved.

		@return: a string containing a date, or None

		@see: L{getBeginDate}
		"""
		return self._endDate

	def setEndDate(self, dateStr):
		"""Sets the death/dissolving date.

		@param dateStr: a string containing a date

		@see: L{setEndDate}, L{getBeginDate}
		"""
		self._endDate = dateStr

	endDate = property(getEndDate, setEndDate,
		doc="The death/dissolving date.")

	def getAliases(self):
		"""Returns the list of aliases for this artist.

		@return: a list of L{ArtistAlias} objects
		"""
		return self._aliases

	aliases = property(getAliases, doc='The list of aliases.')

	def addAlias(self, alias):
		"""Adds an alias for this artist.
		
		@param alias: an L{ArtistAlias} object
		"""
		self._aliases.append(alias)

	def getReleases(self):
		"""Returns a list of releases from this artist.

		This may also include releases where this artist isn't the
		I{main} artist but has just contributed one or more tracks
		(aka VA-Releases).

		@return: a list of L{Release} objects
		"""
		return self._releases

	releases = property(getReleases, doc='The list of releases')

	def addRelease(self, release):
		"""Adds a release to this artist's list of releases.

		@param release: a L{Release} object
		"""
		self._releases.append(release)

	def getReleasesOffset(self):
		"""Returns the offset of the release list.

		This is used if the release list is incomplete (ie. the web
		service only returned part of the release for this artist).
		Note that the offset value is zero-based, which means release
		C{0} is the first release.

		@return: an integer containing the offset, or None

		@see: L{getReleases}, L{getReleasesCount}
		"""
		return self._releasesOffset

	def setReleasesOffset(self, offset):
		"""Sets the offset of the release list.

		@param offset: an integer containing the offset, or None

		@see: L{getReleasesOffset}
		"""
		self._releasesOffset = offset

	releasesOffset = property(getReleasesOffset, setReleasesOffset,
		doc='The offset of the release list.')

	def getReleasesCount(self):
		"""Returns the number of existing releases.
	
		This may or may not match with the number of elements that
		L{getReleases} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setReleasesCount}, L{getReleasesOffset}
		"""
		return self._releasesCount

	def setReleasesCount(self, value):
		"""Sets the number of existing releases.

		@param value: an integer containing the count, or None

		@see: L{getReleasesCount}, L{setReleasesOffset}
		"""
		self._releasesCount = value

	releasesCount = property(getReleasesCount, setReleasesCount,
		doc='The total number of releases')

	def getReleaseGroups(self):
		"""Returns a list of release groups from this artist.
		
		@return: a list of L{ReleaseGroup} objects
		"""
		return self._releaseGroups
	
	releaseGroups = property(getReleaseGroups, doc='The list of release groups')
	
	def addReleaseGroup(self, releaseGroup):
		"""Adds a release group to this artist's list of release groups.
		
		@param releaseGroup: a L{ReleaseGroup} object
		"""
		self._releaseGroups.append(releaseGroup)

	def getReleaseGroupsOffset(self):
		"""Returns the offset of the release group list.

		This is used if the release group list is incomplete (ie. the
		web service only returned part of the result for this artist).
		Note that the offset value is zero-based, which means release
		group C{0} is the first release group.

		@return: an integer containing the offset, or None

		@see: L{getReleaseGroups}, L{getReleaseGroupsCount}
		"""
		return self._releaseGroupsOffset

	def setReleaseGroupsOffset(self, offset):
		"""Sets the offset of the release group list.

		@param offset: an integer containing the offset, or None

		@see: L{getReleaseGroupsOffset}
		"""
		self._releaseGroupsOffset = offset

	releaseGroupsOffset = property(getReleaseGroupsOffset, setReleaseGroupsOffset,
		doc='The offset of the release group list.')

	def getReleaseGroupsCount(self):
		"""Returns the number of existing release groups.

		This may or may not match with the number of elements that
		L{getReleaseGroups} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setReleaseGroupsCount}, L{getReleaseGroupsOffset}
		"""
		return self._releaseGroupsCount

	def setReleaseGroupsCount(self, value):
		"""Sets the number of existing release groups.

		@param value: an integer containing the count, or None

		@see: L{getReleaseGroupsCount}, L{setReleaseGroupsOffset}
		"""
		self._releaseGroupsCount = value

	releasesCount = property(getReleaseGroupsCount, setReleaseGroupsCount,
		doc='The total number of release groups')


class Rating(object):
	"""The representation of a MusicBrain rating.

	The rating can have the following values:

	0 = Unrated
	[1..5] = Rating
	"""
	def __init__(self, value=None, count=None):
		"""Constructor.

		@param value: a string containing the tag's value
		@param count: the number of users who added this tag
		"""
		self._value = value
		self._count = count

	def getValue(self):
		"""Returns a string with the tag's value.

		@return: an integer containing the rating's value, or None
		"""
		return self._value

	def setValue(self, value):
		""" Set the value of this rating.

		0 or None = Clear your rating
		1 - 5 = Rating

		@param value: the rating to apply

		@raise ValueError: if value is not a double or not in the
		range 0 - 5 or None.
		"""
		if value == None:
		  value = 0
		try:
		  value = float(value)
		except ValueError, e:
		  raise ValueError("Value for rating needs to be an" \
		      "float.")
		if value < 0.0 or value > 5.0:
		  raise ValueError("Value needs to be in the range [0..5]")
		self._value = value

	value = property(getValue, setValue, doc='The value of the rating.')

	def getCount(self):
		"""Returns an integer containing the rating's frequency count.

		@return: an integer containing the rating's frequency count,
		or None
		"""
		return self._count
	
	def setCount(self, count):
		"""Sets the frequency count of this rating.
		
		@param count: an integer containing the tag's frequency count
		"""
		self._count = count
		
	count = property(getCount, setCount, doc="This tag's frequency count.")

	def __str__(self):
		return str(self._value)

	def __unicode__(self):
		return unicode(self._value)


class Tag(object):
	"""The representation of a MusicBrainz folksonomy tag.

	The tag's value is the text that's displayed in the tag cloud.
	The count attribute keeps track of how many users added the tag
	to its owning entity.
	"""
	def __init__(self, value=None, count=None):
		"""Constructor.

		@param value: a string containing the tag's value
		@param count: the number of users who added this tag
		"""
		self._value = value
		self._count = count

	def getValue(self):
		"""Returns a string with the tag's value.

		@return: a string containing the tags's value, or None
		"""
		return self._value
	
	def setValue(self, value):
		"""Sets the value of this tag.
		
		@param value: A string containing the value of the tag
		"""
		self._value = value
		
	value = property(getValue, setValue, doc='The value of the text.')

	def getCount(self):
		"""Returns an integer containing the tag's frequency count.

		@return: an integer containing the tags's frequency count, or None
		"""
		return self._count
	
	def setCount(self, count):
		"""Sets the frequency count of this tag.
		
		@param count: an integer containing the tag's frequency count
		"""
		self._count = count
		
	count = property(getCount, setCount, doc="This tag's frequency count.")

	def __str__(self):
		return str(self._value)

	def __unicode__(self):
		return unicode(self._value)


class Label(Entity):
	"""Represents a record label.
	
	A label within MusicBrainz is an L{Entity}. It contains information
	about the label like when it was established, its name, label code and
	other relationships. All release events may be assigned a label.
	"""
	TYPE_UNKNOWN = NS_MMD_1 + 'Unknown'
	
	TYPE_DISTRIBUTOR = NS_MMD_1 + 'Distributor'
	TYPE_HOLDING = NS_MMD_1 + 'Holding'
	TYPE_PRODUCTION = NS_MMD_1 + 'Production'
	
	TYPE_ORIGINAL = NS_MMD_1 + 'OriginalProduction'
	TYPE_BOOTLEG = NS_MMD_1 + 'BootlegProduction'
	TYPE_REISSUE = NS_MMD_1 + 'ReissueProduction'
	
	def __init__(self, id_=None):
		"""Constructor.

		@param id_: a string containing an absolute URI
		"""
		Entity.__init__(self, id_)
		self._type = None
		self._name = None
		self._sortName = None
		self._disambiguation = None
		self._countryId = None
		self._code = None
		self._beginDate = None
		self._endDate = None
		self._aliases = [ ]
	
	def getType(self):
		"""Returns the type of this label.

		@return: a string containing an absolute URI
		"""
		return self._type
	
	def setType(self, type_):
		"""Sets the type of this label.
		
		@param type_: A string containing the absolute URI of the type of label.
		"""
		self._type = type_
	
	type = property(getType, setType, doc='The type of label')
	
	def getName(self):
		"""Returns a string with the name of the label.

		@return: a string containing the label's name, or None
		"""
		return self._name
	
	def setName(self, name):
		"""Sets the name of this label.
		
		@param name: A string containing the name of the label
		"""
		self._name = name
		
	name = property(getName, setName, doc='The name of the label.')
	
	def getSortName(self):
		"""Returns the label's sort name.

		The sort name is the label's name in a special format which
		is better suited for lexicographic sorting. The MusicBrainz
		style guide specifies this format.

		@see: U{The MusicBrainz Style Guidelines
			<http://musicbrainz.org/style.html>}
		"""
		return self._sortName

	def setSortName(self, sortName):
		"""Sets the label's sort name.

		@param sortName: a string containing the label's sort name

		@see: L{getSortName}
		"""
		self._sortName = sortName

	sortName = property(getSortName, setSortName,
		doc="The label's sort name.")

	def getDisambiguation(self):
		"""Returns the disambiguation attribute.

		This attribute may be used if there is more than one label
		with the same name. In this case, disambiguation attributes
		are added to the labels' names to keep them apart.

		@return: a disambiguation string, or None

		@see: L{getUniqueName}
		"""
		return self._disambiguation

	def setDisambiguation(self, disambiguation):
		"""Sets the disambiguation attribute.

		@param disambiguation: a disambiguation string

		@see: L{getDisambiguation}, L{getUniqueName}
		"""
		self._disambiguation = disambiguation

	disambiguation = property(getDisambiguation, setDisambiguation,
		doc="The disambiguation comment.")

	def getUniqueName(self):
		"""Returns a unique label name (using disambiguation).

		This method returns the label's name together with the
		disambiguation attribute in parenthesis if it exists.

		@return: a string containing the unique name

		@see: L{getDisambiguation}
		"""
		d = self.getDisambiguation() 
		if d is not None and d.strip() != '':
			return '%s (%s)' % (self.getName(), d)
		else: 
			return self.getName()

	def getBeginDate(self):
		"""Returns the date this label was established.
		
		@return: A string contained the start date, or None
		"""
		return self._beginDate
	
	def setBeginDate(self, date):
		"""Set the date this label was established.
		
		@param date: A string in the format of YYYY-MM-DD
		"""
		self._beginDate = date
	
	beginDate = property(getBeginDate, setBeginDate,
		doc='The date this label was established.')

	def getEndDate(self):
		"""Returns the date this label closed.

		The returned date has the format 'YYYY', 'YYYY-MM', or 
		'YYYY-MM-DD', depending on how much detail is known.
		
		@return: A string containing the date, or None
		"""
		return self._endDate
	
	def setEndDate(self, date):
		"""Set the date this label closed.

		The date may have the format 'YYYY', 'YYYY-MM', or 
		'YYYY-MM-DD', depending on how much detail is known.
		
		@param date: A string containing the date, or None
		"""
		self._endDate = date
	
	endDate = property(getEndDate, setEndDate,
		doc='The date this label closed.')
		
	def getCountry(self):
		"""Returns the country the label is located.

		@return: a string containing an ISO-3166 country code, or None

		@see: L{musicbrainz2.utils.getCountryName}
		"""
		return self._countryId

	def setCountry(self, country):
		"""Sets the country the label is located.

		@param country: a string containing an ISO-3166 country code
		"""
		self._countryId = country

	country = property(getCountry, setCountry,
		doc='The country the label is located.')

	def getCode(self):
		"""Returns the label code.

		Label codes have been introduced by the IFPI (International
		Federation of Phonogram and Videogram Industries) to uniquely
		identify record labels. The label code consists of 'LC-' and 4
		figures (currently being extended to 5 figures).

		@return: a string containing the label code, or None
		"""
		return self._code

	def setCode(self, code):
		"""Sets the label code.

		@param code: a string containing the label code
		"""
		self._code = code

	code = property(getCode, setCode,
		doc='The label code.')

	def getAliases(self):
		"""Returns the list of aliases for this label.

		@return: a list of L{LabelAlias} objects
		"""
		return self._aliases

	aliases = property(getAliases, doc='The list of aliases.')

	def addAlias(self, alias):
		"""Adds an alias for this label.
		
		@param alias: a L{LabelAlias} object
		"""
		self._aliases.append(alias)


class Release(Entity):
	"""Represents a Release.

	A release within MusicBrainz is an L{Entity} which contains L{Track}
	objects.  Releases may be of more than one type: There can be albums,
	singles, compilations, live recordings, official releases, bootlegs
	etc.

	@note: The current MusicBrainz server implementation supports only a
	limited set of types.
	"""
	TYPE_NONE = NS_MMD_1 + 'None'
	TYPE_NON_ALBUM_TRACKS = NS_MMD_1 + "NonAlbum Track"

	TYPE_ALBUM = NS_MMD_1 + 'Album'
	TYPE_SINGLE = NS_MMD_1 + 'Single'
	TYPE_EP = NS_MMD_1 + 'EP'
	TYPE_COMPILATION = NS_MMD_1 + 'Compilation'
	TYPE_SOUNDTRACK = NS_MMD_1 + 'Soundtrack'
	TYPE_SPOKENWORD = NS_MMD_1 + 'Spokenword'
	TYPE_INTERVIEW = NS_MMD_1 + 'Interview'
	TYPE_AUDIOBOOK = NS_MMD_1 + 'Audiobook'
	TYPE_LIVE = NS_MMD_1 + 'Live'
	TYPE_REMIX = NS_MMD_1 + 'Remix'
	TYPE_OTHER = NS_MMD_1 + 'Other'

	TYPE_OFFICIAL = NS_MMD_1 + 'Official'
	TYPE_PROMOTION = NS_MMD_1 + 'Promotion'
	TYPE_BOOTLEG = NS_MMD_1 + 'Bootleg'
	TYPE_PSEUDO_RELEASE = NS_MMD_1 + 'Pseudo-Release'

	def __init__(self, id_=None, title=None):
		"""Constructor.

		@param id_: a string containing an absolute URI
		@param title: a string containing the title
		"""
		Entity.__init__(self, id_)
		self._types = [ ]
		self._title = title
		self._textLanguage = None
		self._textScript = None
		self._asin = None
		self._artist = None
		self._releaseEvents = [ ]
		#self._releaseEventsCount = None
		self._releaseGroup = None
		self._discs = [ ]
		#self._discIdsCount = None
		self._tracks = [ ]
		self._tracksOffset = None
		self._tracksCount = None


	def getTypes(self):
		"""Returns the types of this release.

		To test for release types, you can use the constants
		L{TYPE_ALBUM}, L{TYPE_SINGLE}, etc.

		@return: a list of strings containing absolute URIs

		@see: L{musicbrainz2.utils.getReleaseTypeName}
		"""
		return self._types

	types = property(getTypes, doc='The list of types for this release.')

	def addType(self, type_):
		"""Add a type to the list of types.

		@param type_: a string containing absolute URIs

		@see: L{getTypes}
		"""
		self._types.append(type_)

	def getTitle(self):
		"""Returns the release's title.

		@return: a string containing the release's title
		"""
		return self._title

	def setTitle(self, title):
		"""Sets the release's title.

		@param title: a string containing the release's title, or None
		"""
		self._title = title

	title = property(getTitle, setTitle, doc='The title of this release.')

	def getTextLanguage(self):
		"""Returns the language used in release and track titles.

		To represent the language, the ISO-639-2/T standard is used,
		which provides three-letter terminological language codes like
		'ENG', 'DEU', 'JPN', 'KOR', 'ZHO' or 'YID'.

		Note that this refers to release and track I{titles}, not
		lyrics.

		@return: a string containing the language code, or None

		@see: L{musicbrainz2.utils.getLanguageName}
		"""
		return self._textLanguage

	def setTextLanguage(self, language):
		"""Sets the language used in releaes and track titles.

		@param language: a string containing a language code

		@see: L{getTextLanguage}
		"""
		self._textLanguage = language

	textLanguage = property(getTextLanguage, setTextLanguage,
		doc='The language used in release and track titles.')

	def getTextScript(self):
		"""Returns the script used in release and track titles.

		To represent the script, ISO-15924 script codes are used.
		Valid codes are, among others: 'Latn', 'Cyrl', 'Hans', 'Hebr'

		Note that this refers to release and track I{titles}, not
		lyrics.

		@return: a string containing the script code, or None

		@see: L{musicbrainz2.utils.getScriptName}
		"""
		return self._textScript

	def setTextScript(self, script):
		"""Sets the script used in releaes and track titles.

		@param script: a string containing a script code

		@see: L{getTextScript}
		"""
		self._textScript = script

	textScript = property(getTextScript, setTextScript,
		doc='The script used in release and track titles.')

	def getAsin(self):
		"""Returns the amazon shop identifier (ASIN).

		The ASIN is a 10-letter code (except for books) assigned
		by Amazon, which looks like 'B000002IT2' or 'B00006I4YD'.

		@return: a string containing the ASIN, or None
		"""
		return self._asin

	def setAsin(self, asin):
		"""Sets the amazon shop identifier (ASIN).

		@param asin: a string containing the ASIN

		@see: L{getAsin}
		"""
		self._asin = asin

	asin = property(getAsin, setAsin, doc='The amazon shop identifier.')

	def getArtist(self):
		"""Returns the main artist of this release.

		@return: an L{Artist} object, or None
		"""
		return self._artist

	def setArtist(self, artist):
		"""Sets this release's main artist.

		@param artist: an L{Artist} object
		"""
		self._artist = artist

	artist = property(getArtist, setArtist,
		doc='The main artist of this release.')

	def getReleaseGroup(self):
		"""Returns the release group to which this release belongs.
		
		@return: a L{ReleaseGroup} object, or None.
		"""
		return self._releaseGroup

	def setReleaseGroup(self, releaseGroup):
		"""Sets the release's release group.
		
		@param releaseGroup: a L{ReleaseGroup} object, or None.
		"""
		self._releaseGroup = releaseGroup

	releaseGroup = property(getReleaseGroup, setReleaseGroup,
		doc='The release group this release belongs to.')

	def isSingleArtistRelease(self):
		"""Checks if this is a single artist's release.

		Returns C{True} if the release's main artist (L{getArtist}) is
		also the main artist for all of the tracks. This is checked by
		comparing the artist IDs.

		Note that the release's artist has to be set (see L{setArtist})
		for this. The track artists may be unset.

		@return: True, if this is a single artist's release
		"""
		releaseArtist = self.getArtist()
		assert releaseArtist is not None, 'Release Artist may not be None!'
		for track in self.getTracks():
			if track.getArtist() is None:
				continue
			if track.getArtist().getId() != releaseArtist.getId():
				return False

		return True

	def getTracks(self):
		"""Returns the tracks this release contains.

		@return: a list containing L{Track} objects

		@see: L{getTracksOffset}, L{getTracksCount}
		"""
		return self._tracks

	tracks = property(getTracks, doc='The list of tracks.')

	def addTrack(self, track):
		"""Adds a track to this release.

		This appends a track at the end of this release's track list.

		@param track: a L{Track} object
		"""
		self._tracks.append(track)

	def getTracksOffset(self):
		"""Returns the offset of the track list.

		This is used if the track list is incomplete (ie. the web
		service only returned part of the tracks on this release).
		Note that the offset value is zero-based, which means track
		C{0} is the first track.

		@return: an integer containing the offset, or None

		@see: L{getTracks}, L{getTracksCount}
		"""
		return self._tracksOffset

	def setTracksOffset(self, offset):
		"""Sets the offset of the track list.

		@param offset: an integer containing the offset, or None

		@see: L{getTracksOffset}, L{setTracksCount}
		"""
		self._tracksOffset = offset

	tracksOffset = property(getTracksOffset, setTracksOffset,
		doc='The offset of the track list.')

	def getTracksCount(self):
		"""Returns the number of tracks on this release.
	
		This may or may not match with the number of elements that
		L{getTracks} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setTracksCount}, L{getTracks}, L{getTracksOffset}
		"""
		return self._tracksCount

	def setTracksCount(self, value):
		"""Sets the number of tracks on this release.

		@param value: an integer containing the count, or None

		@see: L{getTracksCount}, L{setTracksOffset}
		"""
		self._tracksCount = value

	tracksCount = property(getTracksCount, setTracksCount,
		doc='The total number of releases')


	def getReleaseEvents(self):
		"""Returns the list of release events.

		A L{Release} may contain a list of so-called release events,
		each represented using a L{ReleaseEvent} object. Release
		evens specify where and when this release was, well, released.

		@return: a list of L{ReleaseEvent} objects

		@see: L{getReleaseEventsAsDict}
		"""
		return self._releaseEvents

	releaseEvents = property(getReleaseEvents,
		doc='The list of release events.')

	def addReleaseEvent(self, event):
		"""Adds a release event to this release.

		@param event: a L{ReleaseEvent} object

		@see: L{getReleaseEvents}
		"""
		self._releaseEvents.append(event)

	def getReleaseEventsAsDict(self):
		"""Returns the release events represented as a dict.

		Keys are ISO-3166 country codes like 'DE', 'UK', 'FR' etc.
		Values are dates in 'YYYY', 'YYYY-MM' or 'YYYY-MM-DD' format.

		@return: a dict containing (countryCode, date) entries

		@see: L{getReleaseEvents}, L{musicbrainz2.utils.getCountryName}
		"""
		d = { }
		for event in self.getReleaseEvents():
			d[event.getCountry()] = event.getDate()
		return d

	def getEarliestReleaseDate(self):
		"""Returns the earliest release date.

		This favours complete dates. For example, '2006-09' is
		returned if there is '2000', too. If there is no release
		event associated with this release, None is returned.

		@return: a string containing the date, or None 

		@see: L{getReleaseEvents}, L{getReleaseEventsAsDict}
		"""
		event = self.getEarliestReleaseEvent()

		if event is None:
			return None
		else:
			return event.getDate()

	def getEarliestReleaseEvent(self):
		"""Returns the earliest release event.

		This works like L{getEarliestReleaseDate}, but instead of
		just the date, this returns a L{ReleaseEvent} object.

		@return: a L{ReleaseEvent} object, or None 

		@see: L{getReleaseEvents}, L{getEarliestReleaseDate}
		"""
		dates = [ ]
		for event in self.getReleaseEvents():
			date = event.getDate()
			if len(date) == 10:    # 'YYYY-MM-DD'
				dates.append( (date, event) )
			elif len(date) == 7:   # 'YYYY-MM'
				dates.append( (date + '-99', event) )
			else:
				dates.append( (date + '-99-99', event) )

		dates.sort(lambda x, y: cmp(x[0], y[0]))

		if len(dates) > 0:
			return dates[0][1]
		else:
			return None


	#def getReleaseEventsCount(self):
	#	"""Returns the number of release events.
	#
	#	This may or may not match with the number of elements that
	#	getReleaseEvents() returns. If the count is higher than
	#	the list, it indicates that the list is incomplete.
	#	"""
	#	return self._releaseEventsCount

	#def setReleaseEventsCount(self, value):
	#	self._releaseEventsCount = value

	def getDiscs(self):
		"""Returns the discs associated with this release.

		Discs are currently containers for MusicBrainz DiscIDs.
		Note that under rare circumstances (identical TOCs), a
		DiscID could be associated with more than one release.

		@return: a list of L{Disc} objects
		"""
		return self._discs

	discs = property(getDiscs, doc='The list of associated discs.')

	def addDisc(self, disc):
		"""Adds a disc to this release.

		@param disc: a L{Disc} object
		"""
		self._discs.append(disc)

	#def getDiscIdsCount(self):
	#	return self._discIdsCount

	#def setDiscIdsCount(self, value):
	#	self._discIdsCount = value


class ReleaseGroup(Entity):
	"""Represents a ReleaseGroup.

	A ReleaseGroup in MusicBrainz is an L{Entity} which groups several different
	versions of L{Release} objects (e.g., different editions of the same album).

	@see: L{Release}
	@see: L{Entity}
	"""

	def __init__(self, id_=None, title=None):
		"""Constructor.

		@param id_: a string containing an absolute URI
		@param title: a string containing the title
		"""
		Entity.__init__(self, id_)
		self._title = title
		self._id = id_
		self._type = None
		self._releases = [ ]
		self._artist = None
		self._releasesOffset = 0
		self._releasesCount = 0

	def getType(self):
		"""Returns the type of this release group.

		To test for release types, you can use the constants
		L{Release.TYPE_ALBUM}, L{Release.TYPE_SINGLE}, etc.

		@return: a string containing an absolute URI, or None.

		@see: L{musicbrainz2.utils.getReleaseTypeName}
		"""
		return self._type

	def setType(self, type_):
		"""Sets the type of this release group.

		Use a constant from the L{Release} class, such as
		L{Release.TYPE_ALBUM} or L{Release.TYPE_SINGLE} to
		set the value.

		@param type_: a string containing an absolute URI, or None.
		
		@see: L{musicbrainz2.utils.getReleaseTypeName}
		"""
		self._type = type_

	type = property(getType, setType,
		doc = 'The type of this release group.')

	def getReleases(self):
		"""Gets the releases in this release group.

		@return: a list of L{Release} objects
		@see: L{Release}
		"""
		return self._releases

	releases = property(getReleases,
		doc = 'The list of releases in this release group.')

	def addRelease(self, release):
		"""Adds a L{Release} to this release group.

		@param release: a L{Release} object
		"""
		self._releases.append(release)

	def getReleasesOffset(self):
		"""Returns the offset of the release list.

		This is used if the release list is incomplete (i.e., the web
		service only returned a portion of the releases in this release
		group).

		@return: an integer containing the offset, or None.
		@see: L{getReleases}, L{getReleasesCount}
		"""
		return self._releasesOffset

	def setReleasesOffset(self, offset):
		"""Sets the offset of the release list.

		@param offset: an integer containing the offset, or None.
		@see: L{getReleases}, L{getReleasesOffset}
		"""
		self._releasesOffset = offset

	releasesOffset = property(getReleasesOffset, setReleasesOffset,
		doc='The offset of the release list.')

	def getReleasesCount(self):
		"""Returns the number of releases in this release group.

		This may or may not match the number of elements returned by
		L{getReleases}.  If the count is higher than the length of that
		list, then the list is incomplete.

		@return: an integer containing the count, or None
		@see: L{getReleases}, L{setReleasesCount}, L{getReleasesOffset}
		"""
		return self._releasesCount

	def setReleasesCount(self, value):
		"""Sets the number of releases in this release group.

		@param value: an integer containing the count, or None.
		@see: L{getReleases}, L{getReleasesCount}, L{getReleasesOffset}
		"""
		self._releasesCount = value

	releasesCount = property(getReleasesCount, setReleasesCount,
		doc = 'The total number of releases')

	def getTitle(self):
		"""Returns this release group's title.

		@return: a string containing the release group's title
		"""
		return self._title

	def setTitle(self, title):
		"""Sets the release group's title.

		@param title: a string containing the release group's title.
		"""
		self._title = title

	title = property(getTitle, setTitle,
		doc = 'The title of this release group.')

	def getArtist(self):
		"""Returns the main artist of this release group.

		@return: an L{Artist} object, or None
		"""
		return self._artist

	def setArtist(self, artist):
		"""Sets the release group's main artist.

		@param artist: an L{Artist} object
		"""
		self._artist = artist

	artist = property(getArtist, setArtist,
		doc = 'The main artist of this release group')


class Track(Entity):
	"""Represents a track.

	This class represents a track which may appear on one or more releases.
	A track may be associated with exactly one artist (the I{main} artist).

	Using L{getReleases}, you can find out on which releases this track
	appears. To get the track number, too, use the
	L{Release.getTracksOffset} method.

	@note: Currently, the MusicBrainz server doesn't support tracks to
	       be on more than one release.

	@see: L{Release}, L{Artist}
	"""
	def __init__(self, id_=None, title=None):
		"""Constructor.

		@param id_: a string containing an absolute URI
		@param title: a string containing the title
		"""
		Entity.__init__(self, id_)
		self._title = title
		self._artist = None
		self._duration = None
		self._puids = [ ]
		self._releases = [ ]
		self._isrcs = [ ]

	def getTitle(self):
		"""Returns the track's title.

		The style and format of this attribute is specified by the
		style guide.

		@return: a string containing the title, or None

		@see: U{The MusicBrainz Style Guidelines
			<http://musicbrainz.org/style.html>}
		"""
		return self._title

	def setTitle(self, title):
		"""Sets the track's title.

		@param title: a string containing the title

		@see: L{getTitle}
		"""
		self._title = title

	title = property(getTitle, setTitle, doc="The track's title.")

	def getArtist(self):
		"""Returns the main artist of this track.

		@return: an L{Artist} object, or None
		"""
		return self._artist

	def setArtist(self, artist):
		"""Sets this track's main artist.

		@param artist: an L{Artist} object
		"""
		self._artist = artist

	artist = property(getArtist, setArtist, doc="The track's main artist.")

	def getDuration(self):
		"""Returns the duration of this track in milliseconds.

		@return: an int containing the duration in milliseconds, or None
		"""
		return self._duration

	def setDuration(self, duration):
		"""Sets the duration of this track in milliseconds.

		@param duration: an int containing the duration in milliseconds
		"""
		self._duration = duration

	duration = property(getDuration, setDuration,
		doc='The duration in milliseconds.')

	def getDurationSplit(self):
		"""Returns the duration as a (minutes, seconds) tuple.

		If no duration is set, (0, 0) is returned. Seconds are
		rounded towards the ceiling if at least 500 milliseconds
		are left.

		@return: a (minutes, seconds) tuple, both entries being ints
		"""
		duration = self.getDuration()
		if duration is None:
			return (0, 0)
		else:
			seconds = int( round(duration / 1000.0) )
			return (seconds / 60, seconds % 60)

	def getPuids(self):
		"""Returns the PUIDs associated with this track.

		Please note that a PUID may be associated with more than one
		track.

		@return: a list of strings, each containing one PUID
		"""
		return self._puids

	puids = property(getPuids, doc='The list of associated PUIDs.')

	def addPuid(self, puid):
		"""Add a PUID to this track.

		@param puid: a string containing a PUID
		"""
		self._puids.append(puid)
	
	def getISRCs(self):
		"""Returns the ISRCs associated with this track.

		@return: a list of strings, each containing one ISRC
		"""
		return self._isrcs

	isrcs = property(getISRCs, doc='The list of associated ISRCs')

	def addISRC(self, isrc):
		"""Add a ISRC to this track.

		@param isrc: a string containing an ISRC
		"""
		self._isrcs.append(isrc)

	def getReleases(self):
		"""Returns the list of releases this track appears on.

		@return: a list of L{Release} objects
		"""
		return self._releases

	releases = property(getReleases,
		doc='The releases on which this track appears.')

	def addRelease(self, release):
		"""Add a release on which this track appears.

		@param release: a L{Release} object
		"""
		self._releases.append(release)


class Relation(object):
	"""Represents a relation between two Entities.

	There may be an arbitrary number of relations between all first
	class objects in MusicBrainz. The Relation itself has multiple
	attributes, which may or may not be used for a given relation
	type.

	Note that a L{Relation} object only contains the target but not
	the source end of the relation.

	@todo: Add some examples.

	@cvar TO_ARTIST: Identifies relations linking to an artist.
	@cvar TO_RELEASE: Identifies relations linking to a release.
	@cvar TO_TRACK: Identifies relations linking to a track.
	@cvar TO_URL: Identifies relations linking to an URL.

	@cvar DIR_NONE: Relation reading direction doesn't matter.
	@cvar DIR_FORWARD: Relation reading direction is from source to target.
	@cvar DIR_BACKWARD: Relation reading direction is from target to source.
	@cvar DIR_BOTH: Relation reading direction doesn't matter (no longer used!).
	"""
	# Relation target types
	#
	TO_ARTIST = NS_REL_1 + 'Artist'
	TO_RELEASE = NS_REL_1 + 'Release'
	TO_TRACK = NS_REL_1 + 'Track'
	TO_URL = NS_REL_1 + 'Url'

	# Relation reading directions
	#
	DIR_BOTH = 'both'
	DIR_FORWARD = 'forward'
	DIR_BACKWARD = 'backward'
	DIR_NONE = 'none'

	def __init__(self, relationType=None, targetType=None, targetId=None,
			direction=DIR_NONE, attributes=None,
			beginDate=None, endDate=None, target=None):
		"""Constructor.

		@param relationType: a string containing an absolute URI
		@param targetType: a string containing an absolute URI
		@param targetId: a string containing an absolute URI
		@param direction: one of C{Relation.DIR_FORWARD},
		C{Relation.DIR_BACKWARD}, or C{Relation.DIR_NONE}
		@param attributes: a list of strings containing absolute URIs
		@param beginDate: a string containing a date
		@param endDate: a string containing a date
		@param target: an instance of a subclass of L{Entity}
		"""
		self._relationType = relationType
		self._targetType = targetType
		self._targetId = targetId
		self._direction = direction
		self._beginDate = beginDate
		self._endDate = endDate
		self._target = target
		self._attributes = attributes
		if self._attributes is None:
			self._attributes = [ ]

	def getType(self):
		"""Returns this relation's type.

		@return: a string containing an absolute URI, or None 
		"""
		return self._relationType

	def setType(self, type_):
		"""Sets this relation's type.

		@param type_: a string containing an absolute URI
		"""
		self._relationType = type_

	type = property(getType, setType, doc="The relation's type.")

	def getTargetId(self):
		"""Returns the target's ID.

		This is the ID the relation points to. It is an absolute
		URI, and in case of an URL relation, it is a URL.

		@return: a string containing an absolute URI
		"""
		return self._targetId

	def setTargetId(self, targetId):
		"""Sets the target's ID.

		@param targetId: a string containing an absolute URI

		@see: L{getTargetId}
		"""
		self._targetId = targetId

	targetId = property(getTargetId, setTargetId, doc="The target's ID.")

	def getTargetType(self):
		"""Returns the target's type.

		For MusicBrainz data, the following target types are defined:
		 - artists: L{Relation.TO_ARTIST}
		 - releases: L{Relation.TO_RELEASE}
		 - tracks: L{Relation.TO_TRACK}
		 - urls: L{Relation.TO_URL}

		@return: a string containing an absolute URI
		"""
		return self._targetType

	def setTargetType(self, targetType):
		"""Sets the target's type.

		@param targetType: a string containing an absolute URI

		@see: L{getTargetType}
		"""
		self._targetType = targetType

	targetId = property(getTargetId, setTargetId,
		doc="The type of target this relation points to.")

	def getAttributes(self):
		"""Returns a list of attributes describing this relation.

		The attributes permitted depend on the relation type.

		@return: a list of strings containing absolute URIs
		"""
		return self._attributes

	attributes = property(getAttributes,
		doc='The list of attributes describing this relation.')

	def addAttribute(self, attribute):
		"""Adds an attribute to the list.

		@param attribute: a string containing an absolute URI
		"""
		self._attributes.append(attribute)

	def getBeginDate(self):
		"""Returns the begin date.

		The definition depends on the relation's type. It may for
		example be the day of a marriage or the year an artist
		joined a band. For other relation types this may be
		undefined.

		@return: a string containing a date
		"""
		return self._beginDate

	def setBeginDate(self, dateStr):
		"""Sets the begin date.

		@param dateStr: a string containing a date

		@see: L{getBeginDate}
		"""
		self._beginDate = dateStr

	beginDate = property(getBeginDate, setBeginDate, doc="The begin date.")

	def getEndDate(self):
		"""Returns the end date.

		As with the begin date, the definition depends on the
		relation's type. Depending on the relation type, this may
		or may not be defined.

		@return: a string containing a date

		@see: L{getBeginDate}
		"""
		return self._endDate

	def setEndDate(self, dateStr):
		"""Sets the end date.

		@param dateStr: a string containing a date

		@see: L{getBeginDate}
		"""
		self._endDate = dateStr

	endDate = property(getEndDate, setEndDate, doc="The end date.")

	def getDirection(self):
		"""Returns the reading direction.

		The direction may be one of L{Relation.DIR_FORWARD},
		L{Relation.DIR_BACKWARD}, or L{Relation.DIR_NONE},
		depending on how the relation should be read. For example,
		if direction is L{Relation.DIR_FORWARD} for a cover relation,
		it is read as "X is a cover of Y". For some relations there is
		no reading direction (like marriages) and the web service doesn't
		send a direction. In these cases, the direction is set to
		L{Relation.DIR_NONE}.

		@return: L{Relation.DIR_FORWARD}, L{Relation.DIR_BACKWARD},
		or L{Relation.DIR_NONE}
		"""
		return self._direction

	def setDirection(self, direction):
		"""Sets the reading direction.

		@param direction: L{Relation.DIR_FORWARD},
		L{Relation.DIR_BACKWARD}, or L{Relation.DIR_NONE}

		@see: L{getDirection}
		"""
		self._direction = direction

	direction = property(getDirection, setDirection,
		doc="The reading direction.")

	def getTarget(self):
		"""Returns this relation's target object.

		Note that URL relations never have a target object. Use the
		L{getTargetId} method to get the URL.

		@return: a subclass of L{Entity}, or None
		"""
		return self._target

	def setTarget(self, target):
		"""Sets this relation's target object.

		Note that URL relations never have a target object, they
		are set using L{setTargetId}.

		@param target: a subclass of L{Entity}
		"""
		self._target = target

	target = property(getTarget, setTarget,
		doc="The relation's target object.")


class ReleaseEvent(object):
	"""A release event, indicating where and when a release took place.

	All country codes used must be valid ISO-3166 country codes (i.e. 'DE',
	'UK' or 'FR'). The dates are strings and must have the format 'YYYY',
	'YYYY-MM' or 'YYYY-MM-DD'.

	The format of the release medium is a URI that can be compared to the
	constants on this class (L{FORMAT_CD}, L{FORMAT_DVD} and others).
	"""
	FORMAT_CD = NS_MMD_1 + 'CD'
	FORMAT_DVD = NS_MMD_1 + 'DVD'
	FORMAT_SACD = NS_MMD_1 + 'SACD'
	FORMAT_DUALDISC = NS_MMD_1 + 'DualDisc'
	FORMAT_LASERDISC = NS_MMD_1 + 'LaserDisc'
	FORMAT_MINIDISC = NS_MMD_1 + 'MiniDisc'
	FORMAT_VINYL = NS_MMD_1 + 'Vinyl'
	FORMAT_CASSETTE = NS_MMD_1 + 'Cassette'
	FORMAT_CARTRIDGE = NS_MMD_1 + 'Cartridge'
	FORMAT_REEL_TO_REEL = NS_MMD_1 + 'ReelToReel'
	FORMAT_DAT = NS_MMD_1 + 'DAT'
	FORMAT_DIGITAL = NS_MMD_1 + 'Digital'
	FORMAT_WAX_CYLINDER = NS_MMD_1 + 'WaxCylinder'
	FORMAT_PIANO_ROLL = NS_MMD_1 + 'PianoRoll'
	FORMAT_OTHER = NS_MMD_1 + 'Other'

	def __init__(self, country=None, dateStr=None):
		"""Constructor.

		@param country: a string containing an ISO-3166 country code
		@param dateStr: a string containing a date string
		"""
		self._countryId = country
		self._dateStr = dateStr
		self._catalogNumber = None
		self._barcode = None
		self._label = None
		self._format = None

	def getCountry(self):
		"""Returns the country a release took place.

		@note: Due to a server limitation, the web service does not
		return country IDs for release collection queries. This only
		affects the L{musicbrainz2.webservice.Query.getReleases} query.

		@return: a string containing an ISO-3166 country code, or None

		@see: L{musicbrainz2.utils.getCountryName}
		"""
		return self._countryId

	def setCountry(self, country):
		"""Sets the country a release took place.

		@param country: a string containing an ISO-3166 country code
		"""
		self._countryId = country

	country = property(getCountry, setCountry,
		doc='The country a release took place.')

	def getCatalogNumber(self):
		"""Returns the catalog number of this release event.

		@return: A string containing the catalog number, or None
		"""
		return self._catalogNumber
	
	def setCatalogNumber(self, catalogNumber):
		"""Sets the catalog number of this release event.
		
		@param catalogNumber: A string containing the catalog number
		"""
		self._catalogNumber = catalogNumber
	
	catalogNumber = property(getCatalogNumber, setCatalogNumber,
		doc='The catalog number of the release event')
		
	def getBarcode(self):
		"""Returns the barcode of this release event.

		@return: A string containing the barcode, or None
		"""
		return self._barcode
	
	def setBarcode(self, barcode):
		"""Sets the barcode of this release event.
		
		@param barcode: A string containing the barcode
		"""
		self._barcode = barcode
	
	barcode = property(getBarcode, setBarcode,
		doc='The barcode of the release event')
		
	def getLabel(self):
		"""Returns a L{Label} object for the label associated with this release.
		
		@return: a L{Label} object, or None
		"""
		return self._label

	def setLabel(self, label):
		"""Sets the label of this release event.
		
		@param label: A L{Label} object
		"""
		self._label = label

	label = property(getLabel, setLabel, doc='The label of the release')

	def getDate(self):
		"""Returns the date a release took place.

		@return: a string containing a date
		"""
		return self._dateStr

	def setDate(self, dateStr):
		"""Sets the date a release took place.

		@param dateStr: a string containing a date
		"""
		self._dateStr = dateStr

	date = property(getDate, setDate, doc='The date a release took place.')

	def getFormat(self):
		"""Returns the format of the release medium.

		@return: a string containing a URI, or None
		"""
		return self._format

	def setFormat(self, format):
		"""Sets the format of the release medium.

		@param format: a string containing a URI
		"""
		self._format = format

	format = property(getFormat, setFormat,
		doc='The format of the release medium.')


class CDStub(object):
	"""Represents a CD Stub"""

	def __init__(self, disc):
		"""Constructor.

		@param disc: a L{Disc} object to create this CD Stub from
		"""
		assert isinstance(disc, Disc), 'musicbrainz2.model.Disc expected'
		self._disc = disc
		self._tracks = [ ]
		self._title = ""
		self._artist = ""
		self._barcode = ""
		self._comment = ""

	def setTitle(self, title):
		"""Sets the title of this release.

		@param title: a string containing the title
		"""
		self._title = title

	def getTitle(self):
		"""Returns the title of this release.

		@return: a string containing the title
		"""
		return self._title

	title = property(getTitle, setTitle,
	       doc='The title of the release')

	def setArtist(self, artist):
		"""Sets the artist of this release.

		@param artist: a string containing the artist
		"""
		self._artist = artist

	def getArtist(self):
		"""Returns the artist of this release.

		@return: a string containing the artist
		"""
		return self._artist

	artist = property(getArtist, setArtist,
		doc='The artist of the release')

	def setComment(self, comment):
		"""Sets the comment for this release.

		@param comment: a string containing the comment
		"""
		self._comment = comment

	def getComment(self):
		"""Returns the comment for this release.

		@return: a string containing the comment
		"""
		return self._comment

	comment = property(getComment, setComment,
		 doc='Comment for the release (optional)')

	def setBarcode(self, barcode):
		"""Sets the barcode of this release.

		@param barcode: a string containing the barcode
		"""
		self._barcode = barcode

	def getBarcode(self):
		"""Returns the barcode of this release.

		@return: a string containing the barcode
		"""
		return self._barcode

	barcode = property(getBarcode, setBarcode,
		 doc='Barcode for the release (optional)')

	def addTrack(self, title, artist=''):
		"""Add a track to this release

		@param title: a string containing the title of the track
		@param artist: a string containing the artist of the track, 
		               if different to the album artist
		"""
		self._tracks.append((title, artist))

	def getTracks(self):
		"""Return all the tracks on the release.

		@return: a list of tuples containing (title, artist) pairs
		         for each track
		"""
		return self._tracks

	tracks = property(getTracks, doc='The tracks of the release.')

class Disc(object):
	"""Represents an Audio CD.

	This class represents an Audio CD. A disc can have an ID (the
	MusicBrainz DiscID), which is calculated from the CD's table of
	contents (TOC). There may also be data from the TOC like the length
	of the disc in sectors, as well as position and length of the tracks.

	Note that different TOCs, maybe due to different pressings, lead to
	different DiscIDs. Conversely, if two different discs have the same
	TOC, they also have the same DiscID (which is unlikely but not
	impossible). DiscIDs are always 28 characters long and look like this:
	C{'J68I_CDcUFdCRCIbHSEbTBCbooA-'}. Sometimes they are also referred
	to as CDIndex IDs.

	The L{MusicBrainz web service <musicbrainz2.webservice>} only returns
	the DiscID and the number of sectors. The DiscID calculation function 
	L{musicbrainz2.disc.readDisc}, however, can retrieve the other
	attributes of L{Disc} from an Audio CD in the disc drive.
	"""
	def __init__(self, id_=None):
		"""Constructor.

		@param id_: a string containing a 28-character DiscID 
		"""
		self._id = id_
		self._sectors = None
		self._firstTrackNum = None
		self._lastTrackNum = None
		self._tracks = [ ]

	def getId(self):
		"""Returns the MusicBrainz DiscID.

		@return: a string containing a 28-character DiscID 
		"""
		return self._id

	def setId(self, id_):
		"""Sets the MusicBrainz DiscId.

		@param id_: a string containing a 28-character DiscID
		"""
		self._id = id_

	id = property(getId, setId, doc="The MusicBrainz DiscID.")

	def getSectors(self):
		"""Returns the length of the disc in sectors.

		@return: the length in sectors as an integer, or None
		"""
		return self._sectors

	def setSectors(self, sectors):
		"""Sets the length of the disc in sectors.

		@param sectors: the length in sectors as an integer
		"""
		self._sectors = sectors

	sectors = property(getSectors, setSectors,
		doc="The length of the disc in sectors.")

	def getFirstTrackNum(self):
		"""Returns the number of the first track on this disc.

		@return: an int containing the track number, or None
		"""
		return self._firstTrackNum

	def setFirstTrackNum(self, trackNum):
		"""Sets the number of the first track on this disc.

		@param trackNum: an int containing the track number, or None
		"""
		self._firstTrackNum = trackNum

	firstTrackNum = property(getFirstTrackNum, setFirstTrackNum,
		doc="The number of the first track on this disc.")

	def getLastTrackNum(self):
		"""Returns the number of the last track on this disc.

		@return: an int containing the track number, or None
		"""
		return self._lastTrackNum

	def setLastTrackNum(self, trackNum):
		"""Sets the number of the last track on this disc.

		@param trackNum: an int containing the track number, or None
		"""
		self._lastTrackNum = trackNum

	lastTrackNum = property(getLastTrackNum, setLastTrackNum,
		doc="The number of the last track on this disc.")

	def getTracks(self):
		"""Returns the sector offset and length of this disc.

		This method returns a list of tuples containing the track
		offset and length in sectors for all tracks on this disc.
		The track offset is measured from the beginning of the disc,
		the length is relative to the track's offset. Note that the
		leadout track is I{not} included.

		@return: a list of (offset, length) tuples (values are ints)
		"""
		return self._tracks

	tracks = property(getTracks,
		doc='Sector offset and length of all tracks.')

	def addTrack(self, track):
		"""Adds a track to the list.
		
		This method adds an (offset, length) tuple to the list of
		tracks. The leadout track must I{not} be added. The total
		length of the disc can be set using L{setSectors}.

		@param track: an (offset, length) tuple (values are ints)

		@see: L{getTracks}
		"""
		self._tracks.append(track)


class AbstractAlias(object):
	"""An abstract super class for all alias classes."""
	def __init__(self, value=None, type_=None, script=None):
		"""Constructor.

		@param value: a string containing the alias
		@param type_: a string containing an absolute URI
		@param script: a string containing an ISO-15924 script code
		"""
		self._value = value
		self._type = type_
		self._script = script

	def getValue(self):
		"""Returns the alias.

		@return: a string containing the alias
		"""
		return self._value

	def setValue(self, value):
		"""Sets the alias.

		@param value: a string containing the alias
		"""
		self._value = value

	value = property(getValue, setValue, doc='The alias value.')

	def getType(self):
		"""Returns the alias type.

		@return: a string containing an absolute URI, or None 
		"""
		return self._type

	def setType(self, type_):
		"""Sets the alias type.

		@param type_: a string containing an absolute URI, or None
		"""
		self._type = type_

	type = property(getType, setType, doc='The alias type.')

	def getScript(self):
		"""Returns the alias script.

		@return: a string containing an ISO-15924 script code
		"""
		return self._script

	def setScript(self, script):
		"""Sets the alias script.

		@param script: a string containing an ISO-15924 script code
		"""
		self._script = script

	script = property(getScript, setScript, doc='The alias script.')


class ArtistAlias(AbstractAlias):
	"""Represents an artist alias.

	An alias (the I{alias value}) is a different representation of an
	artist's name. This may be a common misspelling or a transliteration
	(the I{alias type}).

	The I{alias script} is interesting mostly for transliterations and
	indicates which script is used for the alias value. To represent the
	script, ISO-15924 script codes like 'Latn', 'Cyrl', or 'Hebr' are used.
	"""
	pass


class LabelAlias(AbstractAlias):
	"""Represents a label alias.

	An alias (the I{alias value}) is a different representation of a
	label's name. This may be a common misspelling or a transliteration
	(the I{alias type}).

	The I{alias script} is interesting mostly for transliterations and
	indicates which script is used for the alias value. To represent the
	script, ISO-15924 script codes like 'Latn', 'Cyrl', or 'Hebr' are used.
	"""
	pass


class User(object):
	"""Represents a MusicBrainz user."""

	def __init__(self):
		"""Constructor."""
		self._name = None
		self._types = [ ]
		self._showNag = None

	def getName(self):
		"""Returns the user name.

		@return: a string containing the user name
		"""
		return self._name

	def setName(self, name):
		"""Sets the user name.

		@param name: a string containing the user name
		"""
		self._name = name

	name = property(getName, setName, doc='The MusicBrainz user name.')

	def getTypes(self):
		"""Returns the types of this user.

		Most users' type list is empty. Currently, the following types
		are defined:

		 - 'http://musicbrainz.org/ns/ext-1.0#AutoEditor'
		 - 'http://musicbrainz.org/ns/ext-1.0#RelationshipEditor'
		 - 'http://musicbrainz.org/ns/ext-1.0#Bot'
		 - 'http://musicbrainz.org/ns/ext-1.0#NotNaggable'

		@return: a list of strings containing absolute URIs
		"""
		return self._types

	types = property(getTypes, doc="The user's types.")

	def addType(self, type_):
		"""Add a type to the list of types.

		@param type_: a string containing absolute URIs

		@see: L{getTypes}
		"""
		self._types.append(type_)

	def getShowNag(self):
		"""Returns true if a nag screen should be displayed to the user.

		@return: C{True}, C{False}, or None
		"""
		return self._showNag

	def setShowNag(self, value):
		"""Sets the value of the nag screen flag.

		If set to C{True}, 

		@param value: C{True} or C{False}

		@see: L{getShowNag}
		"""
		self._showNag = value

	showNag = property(getShowNag, setShowNag,
		doc='The value of the nag screen flag.')

# EOF



"""Support data for the musicbrainz2 package.

This package is I{not} part of the public API, it has been added to work
around shortcomings in python and may thus be removed at any time.

Please use the L{musicbrainz2.utils} module instead.
"""
__revision__ = '$Id: __init__.py 7386 2006-04-30 11:12:55Z matt $'

# EOF



"""Tests for parsing user elements using MbXmlParser."""
import unittest
from musicbrainz2.wsxml import MbXmlParser, ParseError
from musicbrainz2.model import NS_MMD_1, NS_EXT_1
import StringIO
import os.path

VALID_DATA_DIR = os.path.join('test-data', 'valid')
INVALID_DATA_DIR = os.path.join('test-data', 'invalid')

VALID_USER_DIR = os.path.join(VALID_DATA_DIR, 'user')


class ParseUserTest(unittest.TestCase):

	def __init__(self, name):
		unittest.TestCase.__init__(self, name)


	def testUser(self):
		f = os.path.join(VALID_USER_DIR, 'User_1.xml')
		md = MbXmlParser().parse(f)
		userList = md.getUserList()

		self.assertEquals(len(userList), 1)

		user = userList[0]
		self.failIf( user is None )
		self.assertEquals(user.getName(), 'matt')
		self.assertEquals(user.getShowNag(), False)

		types = user.getTypes()
		self.assertEquals(len(types), 2)
		self.assert_( NS_EXT_1 + 'AutoEditor' in types )
		self.assert_( NS_EXT_1 + 'RelationshipEditor' in types )


# EOF



"""A collection of classes for MusicBrainz.

To get started quickly, have a look at L{webservice.Query} and the examples
there. The source distribution also contains example code you might find
interesting.

This package contains the following modules:

 1. L{model}: The MusicBrainz domain model, containing classes like
    L{Artist <model.Artist>}, L{Release <model.Release>}, or
    L{Track <model.Track>}

 2. L{webservice}: An interface to the MusicBrainz XML web service.

 3. L{wsxml}: A parser for the web service XML format (MMD).

 4. L{disc}: Functions for creating and submitting DiscIDs.

 5. L{utils}: Utilities for working with URIs and other commonly needed tools.

@author: Matthias Friedrich <matt@mafr.de>
"""
__revision__ = '$Id: __init__.py 13329 2011-11-05 08:20:21Z luks $'
__version__ = '0.7.4'

# EOF



"""Tests for subclasses of IIncludes."""
import unittest
from musicbrainz2.model import Release
from musicbrainz2.webservice import (
	ArtistIncludes, ReleaseIncludes, TrackIncludes, LabelIncludes)


class ArtistIncludesTest(unittest.TestCase):
	def testReleases(self):
		inc1 = ArtistIncludes(aliases=True, releaseRelations=True)
		tags1 = inc1.createIncludeTags()
		tags1.sort()
		self.assertEqual(tags1, ['aliases', 'release-rels'])

		inc2 = ArtistIncludes(releaseRelations=True)
		tags2 = inc2.createIncludeTags()
		tags2.sort()
		self.assertNotEqual(tags2, ['aliases', 'release-rels'])

		inc3 = ArtistIncludes(aliases=True,
			releases=(Release.TYPE_ALBUM, Release.TYPE_OFFICIAL))
		tags3 = inc3.createIncludeTags()
		tags3.sort()
		self.assertEqual(tags3, ['aliases', 'sa-Album', 'sa-Official'])

		inc4 = ArtistIncludes(aliases=True, vaReleases=('Bootleg',))
		tags4 = inc4.createIncludeTags()
		tags4.sort()
		self.assertEqual(tags4, ['aliases', 'va-Bootleg'])

		inc5 = ArtistIncludes(aliases=True,
			vaReleases=(Release.TYPE_BOOTLEG,))
		tags5 = inc5.createIncludeTags()
		tags5.sort()
		self.assertEqual(tags5, ['aliases', 'va-Bootleg'])

	def testReleaseGroups(self):
		inc = ArtistIncludes(releaseGroups=True)
		self.assertEqual(inc.createIncludeTags(), ['release-groups'])

	def testTags(self):
		def check(includes_class):
			inc1 = includes_class(tags=True)
			tags1 = inc1.createIncludeTags()
			tags1.sort()
			self.assertEqual(tags1, ['tags'])
		check(ArtistIncludes)
		check(ReleaseIncludes)
		check(TrackIncludes)
		check(LabelIncludes)

class ReleaseIncludesTest(unittest.TestCase):
	# Test that including isrcs in release also pulls in tracks
	def testIsrcs(self):
		inc = ReleaseIncludes(isrcs=True)
		tags = inc.createIncludeTags()
		tags.sort()
		self.assertEqual(tags, ['isrcs', 'tracks'])

	# Test that including labels in release also pulls in release events
	def testReleaseEvents(self):
		inc = ReleaseIncludes(labels=True)
		tags = inc.createIncludeTags()
		tags.sort()
		self.assertEqual(tags, ['labels', 'release-events'])

	def testReleaseGroup(self):
		inc = ReleaseIncludes(releaseGroup=True)
		tags = inc.createIncludeTags()
		self.assertEqual(tags, ['release-groups'])


# EOF



"""Tests for the utils module."""
import unittest
import musicbrainz2.model as m
import musicbrainz2.utils as u

class UtilsTest(unittest.TestCase):
	
	def __init__(self, name):
		unittest.TestCase.__init__(self, name)


	def testExtractUuid(self):
		artistPrefix = 'http://musicbrainz.org/artist/'
		uuid = 'c0b2500e-0cef-4130-869d-732b23ed9df5'
		mbid = artistPrefix + uuid

		self.assertEquals(u.extractUuid(None), None)
		self.assertEquals(u.extractUuid(uuid), uuid)
		self.assertEquals(u.extractUuid(mbid), uuid)
		self.assertEquals(u.extractUuid(mbid, 'artist'), uuid)

		# not correct, but not enough data to catch this
		self.assertEquals(u.extractUuid(uuid, 'release'), uuid)

		self.assertRaises(ValueError, u.extractUuid, mbid, 'release')
		self.assertRaises(ValueError, u.extractUuid, mbid, 'track')
		self.assertRaises(ValueError, u.extractUuid, mbid+'/xy', 'artist')

		invalidId = 'http://example.invalid/' + uuid
		self.assertRaises(ValueError, u.extractUuid, invalidId)


	def testExtractFragment(self):
		fragment = 'Album'
		uri = m.NS_MMD_1 + fragment

		self.assertEquals(u.extractFragment(None), None)
		self.assertEquals(u.extractFragment(fragment), fragment)
		self.assertEquals(u.extractFragment(uri), fragment)
		self.assertEquals(u.extractFragment(uri, m.NS_MMD_1), fragment)

		prefix = 'http://example.invalid/'
		self.assertRaises(ValueError, u.extractFragment, uri, prefix)


	def testExtractEntityType(self):
		prefix = 'http://musicbrainz.org'
		uuid = 'c0b2500e-0cef-4130-869d-732b23ed9df5'

		mbid1 = prefix + '/artist/' + uuid
		self.assertEquals(u.extractEntityType(mbid1), 'artist')

		mbid2 = prefix + '/release/' + uuid
		self.assertEquals(u.extractEntityType(mbid2), 'release')

		mbid3 = prefix + '/track/' + uuid
		self.assertEquals(u.extractEntityType(mbid3), 'track')

		mbid4 = prefix + '/label/' + uuid
		self.assertEquals(u.extractEntityType(mbid4), 'label')

		mbid5 = prefix + '/invalid/' + uuid
		self.assertRaises(ValueError, u.extractEntityType, mbid5)

		self.assertRaises(ValueError, u.extractEntityType, None)
		self.assertRaises(ValueError, u.extractEntityType, uuid)

		invalidUri = 'http://example.invalid/foo'
		self.assertRaises(ValueError, u.extractEntityType, invalidUri)
		

	def testGetCountryName(self):
		self.assertEquals(u.getCountryName('DE'), 'Germany')
		self.assertEquals(u.getCountryName('FR'), 'France')

	def testGetLanguageName(self):
		self.assertEquals(u.getLanguageName('DEU'), 'German')
		self.assertEquals(u.getLanguageName('ENG'), 'English')

	def testGetScriptName(self):
		self.assertEquals(u.getScriptName('Latn'), 'Latin')
		self.assertEquals(u.getScriptName('Cyrl'), 'Cyrillic')

	def testGetReleaseTypeName(self):
		self.assertEquals(u.getReleaseTypeName(m.Release.TYPE_ALBUM),
			'Album')
		self.assertEquals(u.getReleaseTypeName(m.Release.TYPE_COMPILATION), 'Compilation')

# EOF



"""Various utilities to simplify common tasks.

This module contains helper functions to make common tasks easier.

@author: Matthias Friedrich <matt@mafr.de>
"""
__revision__ = '$Id: utils.py 13322 2011-11-03 13:38:06Z luks $'

import re
import urlparse

__all__ = [
	'extractUuid', 'extractFragment', 'extractEntityType',
	'getReleaseTypeName', 'getCountryName', 'getLanguageName',
	'getScriptName',
]


# A pattern to split the path part of an absolute MB URI.
PATH_PATTERN = '^/(artist|release|track|label|release-group)/([^/]*)$'


def extractUuid(uriStr, resType=None):
	"""Extract the UUID part from a MusicBrainz identifier.

	This function takes a MusicBrainz ID (an absolute URI) as the input
	and returns the UUID part of the URI, thus turning it into a relative
	URI. If C{uriStr} is None or a relative URI, then it is returned
	unchanged.

	The C{resType} parameter can be used for error checking. Set it to
	'artist', 'release', or 'track' to make sure C{uriStr} is a
	syntactically valid MusicBrainz identifier of the given resource
	type. If it isn't, a C{ValueError} exception is raised.
	This error checking only works if C{uriStr} is an absolute URI, of
	course.

	Example:

	>>> from musicbrainz2.utils import extractUuid
	>>>  extractUuid('http://musicbrainz.org/artist/c0b2500e-0cef-4130-869d-732b23ed9df5', 'artist')
	'c0b2500e-0cef-4130-869d-732b23ed9df5'
	>>>

	@param uriStr: a string containing a MusicBrainz ID (an URI), or None
	@param resType: a string containing a resource type

	@return: a string containing a relative URI, or None

	@raise ValueError: the given URI is no valid MusicBrainz ID
	"""
	if uriStr is None:
		return None

	(scheme, netloc, path) = urlparse.urlparse(uriStr)[:3]

	if scheme == '':
		return uriStr	# no URI, probably already the UUID

	if scheme != 'http' or netloc != 'musicbrainz.org':
		raise ValueError('%s is no MB ID.' % uriStr)

	m = re.match(PATH_PATTERN, path)

	if m:
		if resType is None:
			return m.group(2)
		else:
			if m.group(1) == resType:
				return m.group(2)
			else:
				raise ValueError('expected "%s" Id' % resType)
	else:
		raise ValueError('%s is no valid MB ID.' % uriStr)


def extractFragment(uriStr, uriPrefix=None):
	"""Extract the fragment part from a URI.

	If C{uriStr} is None or no absolute URI, then it is returned unchanged.

	The C{uriPrefix} parameter can be used for error checking. If C{uriStr}
	is an absolute URI, then the function checks if it starts with
	C{uriPrefix}. If it doesn't, a C{ValueError} exception is raised.

	@param uriStr: a string containing an absolute URI
	@param uriPrefix: a string containing an URI prefix

	@return: a string containing the fragment, or None

	@raise ValueError: the given URI doesn't start with C{uriPrefix}
	"""
	if uriStr is None:
		return None

	(scheme, netloc, path, params, query, frag) = urlparse.urlparse(uriStr)
	if scheme == '':
		return uriStr # this is no URI

	if uriPrefix is None or uriStr.startswith(uriPrefix):
		return frag
	else:
		raise ValueError("prefix doesn't match URI %s" % uriStr)


def extractEntityType(uriStr):
	"""Returns the entity type an entity URI is referring to.

	@param uriStr: a string containing an absolute entity URI

	@return: a string containing 'artist', 'release', 'track', or 'label'

	@raise ValueError: if the given URI is no valid MusicBrainz ID
	"""
	if uriStr is None:
		raise ValueError('None is no valid entity URI')

	(scheme, netloc, path) = urlparse.urlparse(uriStr)[:3]

	if scheme == '':
		raise ValueError('%s is no absolute MB ID.' % uriStr)

	if scheme != 'http' or netloc != 'musicbrainz.org':
		raise ValueError('%s is no MB ID.' % uriStr)

	m = re.match(PATH_PATTERN, path)

	if m:
		return m.group(1)
	else:
		raise ValueError('%s is no valid MB ID.' % uriStr)


def getReleaseTypeName(releaseType):
	"""Returns the name of a release type URI.

	@param releaseType: a string containing a release type URI

	@return: a string containing a printable name for the release type

	@see: L{musicbrainz2.model.Release}
	"""
	from musicbrainz2.data.releasetypenames import releaseTypeNames
	return releaseTypeNames.get(releaseType)


def getCountryName(id_):
	"""Returns a country's name based on an ISO-3166 country code.

	The country table this function is based on has been modified for
	MusicBrainz purposes by using the extension mechanism defined in
	ISO-3166. All IDs are still valid ISO-3166 country codes, but some
	IDs have been added to include historic countries and some of the
	country names have been modified to make them better suited for
	display purposes.

	If the country ID is not found, None is returned. This may happen
	for example, when new countries are added to the MusicBrainz web
	service which aren't known to this library yet.

	@param id_: a two-letter upper case string containing an ISO-3166 code

	@return: a string containing the country's name, or None

	@see: L{musicbrainz2.model}
	"""
	from musicbrainz2.data.countrynames import countryNames
	return countryNames.get(id_)


def getLanguageName(id_):
	"""Returns a language name based on an ISO-639-2/T code.

	This function uses a subset of the ISO-639-2/T code table to map
	language IDs (terminologic, not bibliographic ones!) to names.

	@param id_: a three-letter upper case string containing an ISO-639-2/T code

	@return: a string containing the language's name, or None

	@see: L{musicbrainz2.model}
	"""
	from musicbrainz2.data.languagenames import languageNames
	return languageNames.get(id_)


def getScriptName(id_):
	"""Returns a script name based on an ISO-15924 code.

	This function uses a subset of the ISO-15924 code table to map
	script IDs to names.

	@param id_: a four-letter string containing an ISO-15924 script code

	@return: a string containing the script's name, or None

	@see: L{musicbrainz2.model}
	"""
	from musicbrainz2.data.scriptnames import scriptNames
	return scriptNames.get(id_)


# EOF



"""A parser for the Music Metadata XML Format (MMD).

This module contains L{MbXmlParser}, which parses the U{Music Metadata XML
Format (MMD) <http://musicbrainz.org/development/mmd/>} returned by the
MusicBrainz webservice. 

There are also DOM helper functions in this module used by the parser which
probably aren't useful to users.
"""
__revision__ = '$Id: wsxml.py 12028 2009-09-01 13:15:50Z matt $'

import re
import logging
import urlparse
import xml.dom.minidom
import xml.sax.saxutils as saxutils 
from xml.parsers.expat import ExpatError
from xml.dom import DOMException

import musicbrainz2.utils as mbutils
import musicbrainz2.model as model
from musicbrainz2.model import NS_MMD_1, NS_REL_1, NS_EXT_1

__all__ = [
	'DefaultFactory', 'Metadata', 'ParseError',
	'MbXmlParser', 'MbXmlWriter',
	'AbstractResult',
	'ArtistResult', 'ReleaseResult', 'TrackResult', 'LabelResult',
	'ReleaseGroupResult'
]


class DefaultFactory(object):
	"""A factory to instantiate classes from the domain model. 

	This factory may be used to create objects from L{musicbrainz2.model}.
	"""
	def newArtist(self): return model.Artist()
	def newRelease(self): return model.Release()
	def newReleaseGroup(self): return model.ReleaseGroup()
	def newTrack(self): return model.Track()
	def newRelation(self): return model.Relation()
	def newReleaseEvent(self): return model.ReleaseEvent()
	def newDisc(self): return model.Disc()
	def newArtistAlias(self): return model.ArtistAlias()
	def newUser(self): return model.User()
	def newLabel(self): return model.Label()
	def newLabelAlias(self): return model.LabelAlias()
	def newTag(self): return model.Tag()
	def newRating(self): return model.Rating()


class ParseError(Exception):
	"""Exception to be thrown if a parse error occurs.

	The C{'msg'} attribute contains a printable error message, C{'reason'}
	is the lower level exception that was raised.
	"""

	def __init__(self, msg='Parse Error', reason=None):
		Exception.__init__(self)
		self.msg = msg
		self.reason = reason

	def __str__(self):
		return self.msg


class Metadata(object):
	"""Represents a parsed Music Metadata XML document.

	The Music Metadata XML format is very flexible and may contain a
	diverse set of data (e.g. an artist, a release and a list of tracks),
	but usually only a small subset is used (either an artist, a release
	or a track, or a lists of objects from one class).

	@see: L{MbXmlParser} for reading, and L{MbXmlWriter} for writing
		Metadata objects
	"""
	def __init__(self):
		self._artist = None
		self._release = None
		self._track = None
		self._label = None
		self._releaseGroup = None
		self._artistResults = [ ]
		self._artistResultsOffset = None
		self._artistResultsCount = None
		self._releaseResults = [ ]
		self._releaseResultsOffset = None
		self._releaseResultsCount = None
		self._releaseGroupResults = [ ]
		self._releaseGroupResultsOffset = None
		self._releaseGroupResultsCount = None
		self._trackResults = [ ]
		self._trackResultsOffset = None
		self._trackResultsCount = None
		self._labelResults = [ ]
		self._labelResultsOffset = None
		self._labelResultsCount = None
		self._tagList = [ ]
		self._rating = None
		self._userList = [ ]

	def getArtist(self):
		return self._artist

	def setArtist(self, artist):
		self._artist = artist

	artist = property(getArtist, setArtist, doc='An Artist object.')
	
	def getLabel(self):
		return self._label
	
	def setLabel(self, label):
		self._label = label
	
	label = property(getLabel, setLabel, doc='A Label object.')

	def getRelease(self):
		return self._release

	def setRelease(self, release):
		self._release = release

	release = property(getRelease, setRelease, doc='A Release object.')

	def getReleaseGroup(self):
		return self._releaseGroup

	def setReleaseGroup(self, releaseGroup):
		self._releaseGroup = releaseGroup

	releaseGroup = property(getReleaseGroup, setReleaseGroup)

	def getTrack(self):
		return self._track

	def setTrack(self, track):
		self._track = track

	track = property(getTrack, setTrack, doc='A Track object.')

	def getArtistResults(self):
		"""Returns an artist result list. 

		@return: a list of L{ArtistResult} objects.
		"""
		return self._artistResults

	artistResults = property(getArtistResults,
		doc='A list of ArtistResult objects.')

	def getArtistResultsOffset(self):
		"""Returns the offset of the artist result list.

		The offset is used for paging through the result list. It
		is zero-based.

		@return: an integer containing the offset, or None

		@see: L{getArtistResults}, L{getArtistResultsCount}
		"""
		return self._artistResultsOffset

	def setArtistResultsOffset(self, value):
		"""Sets the offset of the artist result list.

		@param value: an integer containing the offset, or None

		@see: L{getArtistResultsOffset}
		"""
		self._artistResultsOffset = value

	artistResultsOffset = property(
		getArtistResultsOffset, setArtistResultsOffset,
		doc='The offset of the artist results.')

	def getArtistResultsCount(self):
		"""Returns the total number of results available.
	
		This may or may not match with the number of elements that
		L{getArtistResults} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setArtistResultsCount}, L{getArtistResultsOffset}
		"""
		return self._artistResultsCount

	def setArtistResultsCount(self, value):
		"""Sets the total number of available results.

		@param value: an integer containing the count, or None

		@see: L{getArtistResults}, L{setArtistResultsOffset}
		"""
		self._artistResultsCount = value

	artistResultsCount = property(
		getArtistResultsCount, setArtistResultsCount,
		doc='The total number of artists results.')

	def getLabelResults(self):
		"""Returns a label result list.
		
		@return: a list of L{LabelResult} objects.
		"""
		return self._labelResults
	
	labelResults = property(getLabelResults,
		doc='A list of LabelResult objects')

	def getLabelResultsOffset(self):
		"""Returns the offset of the label result list.

		The offset is used for paging through the result list. It
		is zero-based.

		@return: an integer containing the offset, or None

		@see: L{getLabelResults}, L{getLabelResultsCount}
		"""
		return self._labelResultsOffset

	def setLabelResultsOffset(self, value):
		"""Sets the offset of the label result list.

		@param value: an integer containing the offset, or None

		@see: L{getLabelResultsOffset}
		"""
		self._labelResultsOffset = value

	labelResultsOffset = property(
		getLabelResultsOffset, setLabelResultsOffset,
		doc='The offset of the label results.')

	def getLabelResultsCount(self):
		"""Returns the total number of results available.
	
		This may or may not match with the number of elements that
		L{getLabelResults} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setLabelResultsCount}, L{getLabelResultsOffset}
		"""
		return self._labelResultsCount

	def setLabelResultsCount(self, value):
		"""Sets the total number of available results.

		@param value: an integer containing the count, or None

		@see: L{getLabelResults}, L{setLabelResultsOffset}
		"""
		self._labelResultsCount = value

	labelResultsCount = property(
		getLabelResultsCount, setLabelResultsCount,
		doc='The total number of label results.')

	def getReleaseResults(self):
		"""Returns a release result list. 

		@return: a list of L{ReleaseResult} objects.
		"""
		return self._releaseResults

	releaseResults = property(getReleaseResults,
		doc='A list of ReleaseResult objects.')

	def getReleaseResultsOffset(self):
		"""Returns the offset of the release result list.

		The offset is used for paging through the result list. It
		is zero-based.

		@return: an integer containing the offset, or None

		@see: L{getReleaseResults}, L{getReleaseResultsCount}
		"""
		return self._releaseResultsOffset

	def setReleaseResultsOffset(self, value):
		"""Sets the offset of the release result list.

		@param value: an integer containing the offset, or None

		@see: L{getReleaseResultsOffset}
		"""
		self._releaseResultsOffset = value

	releaseResultsOffset = property(
		getReleaseResultsOffset, setReleaseResultsOffset,
		doc='The offset of the release results.')

	def getReleaseResultsCount(self):
		"""Returns the total number of results available.
	
		This may or may not match with the number of elements that
		L{getReleaseResults} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setReleaseResultsCount}, L{getReleaseResultsOffset}
		"""
		return self._releaseResultsCount

	def setReleaseResultsCount(self, value):
		"""Sets the total number of available results.

		@param value: an integer containing the count, or None

		@see: L{getReleaseResults}, L{setReleaseResultsOffset}
		"""
		self._releaseResultsCount = value

	releaseResultsCount = property(
		getReleaseResultsCount, setReleaseResultsCount,
		doc='The total number of release results.')

	def getReleaseGroupResults(self):
		"""Returns a release group result list.

		@return: a list of L{ReleaseGroupResult} objects.
		"""
		return self._releaseGroupResults

	releaseGroupResults = property(getReleaseGroupResults,
		doc = 'A list of ReleaseGroupResult objects.')

	def getReleaseGroupResultsOffset(self):
		"""Returns the offset of the release group result list.

		The offset is used for paging through the result list.  It
		is zero-based.

		@return: an integer containing the offset, or None.

		@see: L{getReleaseGroupResults}, L{getReleaseGroupResultsCount}
		"""
		return self._releaseGroupResultsOffset

	def setReleaseGroupResultsOffset(self, value):
		"""Sets the offset of the release group result list.

		@param value: an integer containing the offset, or None

		@see: L{getReleaseGroupResultsOffset}
		"""
		self._releaseGroupResultsOffset = value

	releaseGroupResultsOffset = property(
		getReleaseGroupResultsOffset, setReleaseGroupResultsOffset,
		doc='The offset of the release group results.')

	def getReleaseGroupResultsCount(self):
		"""Returns the total number of results available.

		This may or may not match with the number of elements that
		L{getReleaseGroupResults} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setReleaseGroupResultsCount}, L{getReleaseGroupResultsOffset}
		"""
		return self._releaseGroupResultsCount

	def setReleaseGroupResultsCount(self, value):
		"""Sets the total number of available results.

		@param value: an integer containing the count, or None

		@see: L{getReleaseGroupResults}, L{setReleaseGroupResultsOffset}
		"""
		self._releaseGroupResultsCount = value

	releaseGroupResultsCount = property(
		getReleaseGroupResultsCount, setReleaseGroupResultsCount,
		doc='The total number of release group results.')

	def getTrackResults(self):
		"""Returns a track result list. 

		@return: a list of L{TrackResult} objects.
		"""
		return self._trackResults

	trackResults = property(getTrackResults,
		doc='A list of TrackResult objects.')

	def getTrackResultsOffset(self):
		"""Returns the offset of the track result list.

		The offset is used for paging through the result list. It
		is zero-based.

		@return: an integer containing the offset, or None

		@see: L{getTrackResults}, L{getTrackResultsCount}
		"""
		return self._trackResultsOffset

	def setTrackResultsOffset(self, value):
		"""Sets the offset of the track result list.

		@param value: an integer containing the offset, or None

		@see: L{getTrackResultsOffset}
		"""
		self._trackResultsOffset = value

	trackResultsOffset = property(
		getTrackResultsOffset, setTrackResultsOffset,
		doc='The offset of the track results.')

	def getTrackResultsCount(self):
		"""Returns the total number of results available.
	
		This may or may not match with the number of elements that
		L{getTrackResults} returns. If the count is higher than
		the list, it indicates that the list is incomplete.

		@return: an integer containing the count, or None

		@see: L{setTrackResultsCount}, L{getTrackResultsOffset}
		"""
		return self._trackResultsCount

	def setTrackResultsCount(self, value):
		"""Sets the total number of available results.

		@param value: an integer containing the count, or None

		@see: L{getTrackResults}, L{setTrackResultsOffset}
		"""
		self._trackResultsCount = value

	trackResultsCount = property(
		getTrackResultsCount, setTrackResultsCount,
		doc='The total number of track results.')


	def getTagList(self):
		"""Returns a list of tags.

		@return: a list of L{model.Tag} objects
		"""
		return self._tagList

	tagResults = property(getTagList,
		doc='A list of Tag objects.')

	def getRating(self):
		"""Returns the rating.

		@return: rating object
		"""
		return self._rating

	def setRating(self, value):
		"""Sets the rating.

		@param value: a L{model.Rating} object
		"""
		self._rating = value

	rating = property(getRating, setRating, doc='A Rating object.')


	# MusicBrainz extension to the schema
	def getUserList(self):
		"""Returns a list of users.

		@return: a list of L{model.User} objects

		@note: This is a MusicBrainz extension.
		"""
		return self._userList

	userResults = property(getUserList,
		doc='A list of User objects.')


class AbstractResult(object):
	"""The abstract representation of a result.

	A result is an instance of some kind (Artist, Release, ...)
	associated with a score.
	"""

	def __init__(self, score):
		self._score = score

	def getScore(self):
		"""Returns the result score.

		The score indicates how good this result matches the search
		parameters. The higher the value, the better the match.

		@return: an int between 0 and 100 (both inclusive), or None
		"""
		return self._score

	def setScore(self, score):
		self._score = score

	score = property(getScore, setScore, doc='The relevance score.')


class ArtistResult(AbstractResult):
	"""Represents an artist result.

	An ArtistResult consists of a I{score} and an artist. The score is a
	number between 0 and 100, where a higher number indicates a better
	match.
	"""
	def __init__(self, artist, score):
		super(ArtistResult, self).__init__(score)
		self._artist = artist

	def getArtist(self):
		"""Returns an Artist object.

		@return: a L{musicbrainz2.model.Artist} object
		"""
		return self._artist

	def setArtist(self, artist):
		self._artist = artist

	artist = property(getArtist, setArtist, doc='An Artist object.')


class ReleaseResult(AbstractResult):
	"""Represents a release result.

	A ReleaseResult consists of a I{score} and a release. The score is a
	number between 0 and 100, where a higher number indicates a better
	match.
	"""
	def __init__(self, release, score):
		super(ReleaseResult, self).__init__(score)
		self._release = release

	def getRelease(self):
		"""Returns a Release object.

		@return: a L{musicbrainz2.model.Release} object
		"""
		return self._release

	def setRelease(self, release):
		self._release = release

	release = property(getRelease, setRelease, doc='A Release object.')

class ReleaseGroupResult(AbstractResult):
	"""Represents a release group result.

	A ReleaseGroupResult consists of a I{score} and a release group.  The
	score is a number between 0 and 100, where a higher number indicates
	a better match.
	"""
	def __init__(self, releaseGroup, score):
		super(ReleaseGroupResult, self).__init__(score)
		self._releaseGroup = releaseGroup

	def getReleaseGroup(self):
		"""Returns a ReleaseGroup object.

		@return: a L{musicbrainz2.model.ReleaseGroup} object
		"""
		return self._releaseGroup

	def setReleaseGroup(self, value):
		self._releaseGroup = value

	releaseGroup = property(getReleaseGroup, setReleaseGroup, doc='A ReleaseGroup object.')

class TrackResult(AbstractResult):
	"""Represents a track result.

	A TrackResult consists of a I{score} and a track. The score is a
	number between 0 and 100, where a higher number indicates a better
	match.
	"""
	def __init__(self, track, score):
		super(TrackResult, self).__init__(score)
		self._track = track

	def getTrack(self):
		"""Returns a Track object.

		@return: a L{musicbrainz2.model.Track} object
		"""
		return self._track

	def setTrack(self, track):
		self._track = track

	track = property(getTrack, setTrack, doc='A Track object.')


class LabelResult(AbstractResult):
	"""Represents a label result.

	An LabelResult consists of a I{score} and a label. The score is a
	number between 0 and 100, where a higher number indicates a better
	match.
	"""
	def __init__(self, label, score):
		super(LabelResult, self).__init__(score)
		self._label = label

	def getLabel(self):
		"""Returns a Label object.

		@return: a L{musicbrainz2.model.Label} object
		"""
		return self._label

	def setLabel(self, label):
		self._label = label

	label = property(getLabel, setLabel, doc='A Label object.')


class MbXmlParser(object):
	"""A parser for the Music Metadata XML format.

	This parser supports all basic features and extensions defined by
	MusicBrainz, including unlimited document nesting. By default it
	reads an XML document from a file-like object (stream) and returns
	an object tree representing the document using classes from
	L{musicbrainz2.model}.

	The implementation tries to be as permissive as possible. Invalid
	contents are skipped, but documents have to be well-formed and using
	the correct namespace. In case of unrecoverable errors, a L{ParseError}
	exception is raised.

	@see: U{The Music Metadata XML Format
		<http://musicbrainz.org/development/mmd/>}
	"""

	def __init__(self, factory=DefaultFactory()):
		"""Constructor.

		The C{factory} parameter has be an instance of L{DefaultFactory}
		or a subclass of it. It is used by L{parse} to obtain objects
		from L{musicbrainz2.model} to build resulting object tree.
		If you supply your own factory, you have to make sure all
		returned objects have the same interface as their counterparts
		from L{musicbrainz2.model}.

		@param factory: an object factory 
		"""
		self._log = logging.getLogger(str(self.__class__))
		self._factory = factory

	def parse(self, inStream):
		"""Parses the MusicBrainz web service XML.

		Returns a L{Metadata} object representing the parsed XML or
		raises a L{ParseError} exception if the data was malformed.
		The parser tries to be liberal and skips invalid content if
		possible.

		Note that an L{IOError} may be raised if there is a problem
		reading C{inStream}.

		@param inStream: a file-like object
		@return: a L{Metadata} object (never None)
		@raise ParseError: if the document is not valid
		@raise IOError: if reading from the stream failed
		"""

		try:
			doc = xml.dom.minidom.parse(inStream)

			# Try to find the root element. If this isn't an mmd
			# XML file or the namespace is wrong, this will fail.
			elems = doc.getElementsByTagNameNS(NS_MMD_1, 'metadata')

			if len(elems) != 0:
				md = self._createMetadata(elems[0])
			else:
				msg = 'cannot find root element mmd:metadata'
				self._log.debug('ParseError: ' + msg)
				raise ParseError(msg)

			doc.unlink()

			return md
		except ExpatError, e:
			self._log.debug('ExpatError: ' + str(e))
			raise ParseError(msg=str(e), reason=e)
		except DOMException, e:
			self._log.debug('DOMException: ' + str(e))
			raise ParseError(msg=str(e), reason=e)
			

	def _createMetadata(self, metadata):
		md = Metadata()

		for node in _getChildElements(metadata):
			if _matches(node, 'artist'):
				md.artist = self._createArtist(node)
			elif _matches(node, 'release'):
				md.release = self._createRelease(node)
			elif _matches(node, 'release-group'):
				md.releaseGroup = self._createReleaseGroup(node)
			elif _matches(node, 'track'):
				md.track = self._createTrack(node)
			elif _matches(node, 'label'):
				md.label = self._createLabel(node)
			elif _matches(node, 'artist-list'):
				(offset, count) = self._getListAttrs(node)
				md.artistResultsOffset = offset
				md.artistResultsCount = count
				self._addArtistResults(node, md.getArtistResults())
			elif _matches(node, 'release-list'):
				(offset, count) = self._getListAttrs(node)
				md.releaseResultsOffset = offset
				md.releaseResultsCount = count
				self._addReleaseResults(node, md.getReleaseResults())
			elif _matches(node, 'release-group-list'):
				(offset, count) = self._getListAttrs(node)
				md.releaseGroupResultsOffset = offset
				md.releaseGroupResultsCount = count
				self._addReleaseGroupResults(node, md.getReleaseGroupResults())
			elif _matches(node, 'track-list'):
				(offset, count) = self._getListAttrs(node)
				md.trackResultsOffset = offset
				md.trackResultsCount = count
				self._addTrackResults(node, md.getTrackResults())
			elif _matches(node, 'label-list'):
				(offset, count) = self._getListAttrs(node)
				md.labelResultsOffset = offset
				md.labelResultsCount = count
				self._addLabelResults(node, md.getLabelResults())
			elif _matches(node, 'tag-list'):
				self._addTagsToList(node, md.getTagList())
			elif _matches(node, 'user-list', NS_EXT_1):
				self._addUsersToList(node, md.getUserList())

		return md


	def _addArtistResults(self, listNode, resultList):
		for c in _getChildElements(listNode):
			artist = self._createArtist(c)
			score = _getIntAttr(c, 'score', 0, 100, ns=NS_EXT_1)
			if artist is not None:
				resultList.append(ArtistResult(artist, score))

	def _addReleaseResults(self, listNode, resultList):
		for c in _getChildElements(listNode):
			release = self._createRelease(c)
			score = _getIntAttr(c, 'score', 0, 100, ns=NS_EXT_1)
			if release is not None:
				resultList.append(ReleaseResult(release, score))

	def _addReleaseGroupResults(self, listNode, resultList):
		for c in _getChildElements(listNode):
			releaseGroup = self._createReleaseGroup(c)
			score = _getIntAttr(c, 'score', 0, 100, ns=NS_EXT_1)
			if releaseGroup is not None:
				resultList.append(ReleaseGroupResult(releaseGroup, score))

	def _addTrackResults(self, listNode, resultList):
		for c in _getChildElements(listNode):
			track = self._createTrack(c)
			score = _getIntAttr(c, 'score', 0, 100, ns=NS_EXT_1)
			if track is not None:
				resultList.append(TrackResult(track, score))

	def _addLabelResults(self, listNode, resultList):
		for c in _getChildElements(listNode):
			label = self._createLabel(c)
			score = _getIntAttr(c, 'score', 0, 100, ns=NS_EXT_1)
			if label is not None:
				resultList.append(LabelResult(label, score))

	def _addReleasesToList(self, listNode, resultList):
		self._addToList(listNode, resultList, self._createRelease)

	def _addReleaseGroupsToList(self, listNode, resultList):
		self._addToList(listNode, resultList, self._createReleaseGroup)

	def _addTracksToList(self, listNode, resultList):
		self._addToList(listNode, resultList, self._createTrack)

	def _addUsersToList(self, listNode, resultList):
		self._addToList(listNode, resultList, self._createUser)

	def _addTagsToList(self, listNode, resultList):
		self._addToList(listNode, resultList, self._createTag)

	def _addTagsToEntity(self, listNode, entity):
		for node in _getChildElements(listNode):
			tag = self._createTag(node)
			entity.addTag(tag)

	def _addRatingToEntity(self, attrNode, entity):
		rating = self._createRating(attrNode)
		entity.setRating(rating)

	def _addToList(self, listNode, resultList, creator):
		for c in _getChildElements(listNode):
			resultList.append(creator(c))

	def _getListAttrs(self, listNode):
		offset = _getIntAttr(listNode, 'offset')
		count = _getIntAttr(listNode, 'count')
		return (offset, count)


	def _createArtist(self, artistNode):
		artist = self._factory.newArtist()
		artist.setId(_getIdAttr(artistNode, 'id', 'artist'))
		artist.setType(_getUriAttr(artistNode, 'type'))
		
		for node in _getChildElements(artistNode):
			if _matches(node, 'name'):
				artist.setName(_getText(node))
			elif _matches(node, 'sort-name'):
				artist.setSortName(_getText(node))
			elif _matches(node, 'disambiguation'):
				artist.setDisambiguation(_getText(node))
			elif _matches(node, 'life-span'):
				artist.setBeginDate(_getDateAttr(node, 'begin'))
				artist.setEndDate(_getDateAttr(node, 'end'))
			elif _matches(node, 'alias-list'):
				self._addArtistAliases(node, artist)
			elif _matches(node, 'release-list'):
				(offset, count) = self._getListAttrs(node)
				artist.setReleasesOffset(offset)
				artist.setReleasesCount(count)
				self._addReleasesToList(node, artist.getReleases())
			elif _matches(node, 'release-group-list'):
				(offset, count) = self._getListAttrs(node)
				artist.setReleaseGroupsOffset(offset)
				artist.setReleaseGroupsCount(count)
				self._addReleaseGroupsToList(node, artist.getReleaseGroups())
			elif _matches(node, 'relation-list'):
				self._addRelationsToEntity(node, artist)
			elif _matches(node, 'tag-list'):
				self._addTagsToEntity(node, artist)
			elif _matches(node, 'rating'):
				self._addRatingToEntity(node, artist)

		return artist

	def _createLabel(self, labelNode):
		label = self._factory.newLabel()
		label.setId(_getIdAttr(labelNode, 'id', 'label'))
		label.setType(_getUriAttr(labelNode, 'type'))
		
		for node in _getChildElements(labelNode):
			if _matches(node, 'name'):
				label.setName(_getText(node))
			if _matches(node, 'sort-name'):
				label.setSortName(_getText(node))
			elif _matches(node, 'disambiguation'):
				label.setDisambiguation(_getText(node))
			elif _matches(node, 'label-code'):
				label.setCode(_getText(node))
			elif _matches(node, 'country'):
				country = _getText(node, '^[A-Z]{2}$')
				label.setCountry(country)
			elif _matches(node, 'life-span'):
				label.setBeginDate(_getDateAttr(node, 'begin'))
				label.setEndDate(_getDateAttr(node, 'end'))
			elif _matches(node, 'alias-list'):
				self._addLabelAliases(node, label)
			elif _matches(node, 'tag-list'):
				self._addTagsToEntity(node, label)
			elif _matches(node, 'rating'):
				self._addRatingToEntity(node, label)

		return label

	def _createRelease(self, releaseNode):
		release = self._factory.newRelease()
		release.setId(_getIdAttr(releaseNode, 'id', 'release'))
		for t in _getUriListAttr(releaseNode, 'type'):
			release.addType(t)

		for node in _getChildElements(releaseNode):
			if _matches(node, 'title'):
				release.setTitle(_getText(node))
			elif _matches(node, 'text-representation'):
				lang = _getAttr(node, 'language', '^[A-Z]{3}$')
				release.setTextLanguage(lang)
				script = _getAttr(node, 'script', '^[A-Z][a-z]{3}$')
				release.setTextScript(script)
			elif _matches(node, 'asin'):
				release.setAsin(_getText(node))
			elif _matches(node, 'artist'):
				release.setArtist(self._createArtist(node))
			elif _matches(node, 'release-event-list'):
				self._addReleaseEvents(node, release)
			elif _matches(node, 'release-group'):
				release.setReleaseGroup(self._createReleaseGroup(node))
			elif _matches(node, 'disc-list'):
				self._addDiscs(node, release)
			elif _matches(node, 'track-list'):
				(offset, count) = self._getListAttrs(node)
				release.setTracksOffset(offset)
				release.setTracksCount(count)
				self._addTracksToList(node, release.getTracks())
			elif _matches(node, 'relation-list'):
				self._addRelationsToEntity(node, release)
			elif _matches(node, 'tag-list'):
				self._addTagsToEntity(node, release)
			elif _matches(node, 'rating'):
				self._addRatingToEntity(node, release)

		return release

	def _createReleaseGroup(self, node):
		rg = self._factory.newReleaseGroup()
		rg.setId(_getIdAttr(node, 'id', 'release-group'))
		rg.setType(_getUriAttr(node, 'type'))

		for child in _getChildElements(node):
			if _matches(child, 'title'):
				rg.setTitle(_getText(child))
			elif _matches(child, 'artist'):
				rg.setArtist(self._createArtist(child))
			elif _matches(child, 'release-list'):
				(offset, count) = self._getListAttrs(child)
				rg.setReleasesOffset(offset)
				rg.setReleasesCount(count)
				self._addReleasesToList(child, rg.getReleases())

		return rg

	def _addReleaseEvents(self, releaseListNode, release):
		for node in _getChildElements(releaseListNode):
			if _matches(node, 'event'):
				country = _getAttr(node, 'country', '^[A-Z]{2}$')
				date = _getDateAttr(node, 'date')
				catalogNumber = _getAttr(node, 'catalog-number')
				barcode = _getAttr(node, 'barcode')
				format = _getUriAttr(node, 'format')

				# The date attribute is mandatory. If it isn't present,
				# we don't add anything from this release event.
				if date is not None:
					event = self._factory.newReleaseEvent()
					event.setCountry(country)
					event.setDate(date)
					event.setCatalogNumber(catalogNumber)
					event.setBarcode(barcode)
					event.setFormat(format)
					
					for subNode in _getChildElements(node):
						if _matches(subNode, 'label'):
							event.setLabel(self._createLabel(subNode))
					
					release.addReleaseEvent(event)


	def _addDiscs(self, discIdListNode, release):
		for node in _getChildElements(discIdListNode):
			if _matches(node, 'disc') and node.hasAttribute('id'):
				d = self._factory.newDisc()
				d.setId(node.getAttribute('id'))
				d.setSectors(_getIntAttr(node, 'sectors', 0))
				release.addDisc(d)


	def _addArtistAliases(self, aliasListNode, artist):
		for node in _getChildElements(aliasListNode):
			if _matches(node, 'alias'):
				alias = self._factory.newArtistAlias()
				self._initializeAlias(alias, node)
				artist.addAlias(alias)


	def _addLabelAliases(self, aliasListNode, label):
		for node in _getChildElements(aliasListNode):
			if _matches(node, 'alias'):
				alias = self._factory.newLabelAlias()
				self._initializeAlias(alias, node)
				label.addAlias(alias)


	def _initializeAlias(self, alias, node):
		alias.setValue(_getText(node))
		alias.setType(_getUriAttr(node, 'type'))
		alias.setScript(_getAttr(node, 'script',
			'^[A-Z][a-z]{3}$'))


	def _createTrack(self, trackNode):
		track = self._factory.newTrack()
		track.setId(_getIdAttr(trackNode, 'id', 'track'))

		for node in _getChildElements(trackNode):
			if _matches(node, 'title'):
				track.setTitle(_getText(node))
			elif _matches(node, 'artist'):
				track.setArtist(self._createArtist(node))
			elif _matches(node, 'duration'):
				track.setDuration(_getPositiveIntText(node))
			elif _matches(node, 'release-list'):
				self._addReleasesToList(node, track.getReleases())
			elif _matches(node, 'puid-list'):
				self._addPuids(node, track)
			elif _matches(node, 'isrc-list'):
				self._addISRCs(node, track)
			elif _matches(node, 'relation-list'):
				self._addRelationsToEntity(node, track)
			elif _matches(node, 'tag-list'):
				self._addTagsToEntity(node, track)
			elif _matches(node, 'rating'):
				self._addRatingToEntity(node, track)

		return track

	# MusicBrainz extension
	def _createUser(self, userNode):
		user = self._factory.newUser()
		for t in _getUriListAttr(userNode, 'type', NS_EXT_1):
			user.addType(t)

		for node in _getChildElements(userNode):
			if _matches(node, 'name'):
				user.setName(_getText(node))
			elif _matches(node, 'nag', NS_EXT_1):
				user.setShowNag(_getBooleanAttr(node, 'show'))

		return user

	def _createRating(self, ratingNode):
		rating = self._factory.newRating()
		rating.value = _getText(ratingNode)
		rating.count = _getIntAttr(ratingNode, 'votes-count')
		return rating

	def _createTag(self, tagNode):
		tag = self._factory.newTag()
		tag.value = _getText(tagNode)
		tag.count = _getIntAttr(tagNode, 'count')
		return tag

	
	def _addPuids(self, puidListNode, track):
		for node in _getChildElements(puidListNode):
			if _matches(node, 'puid') and node.hasAttribute('id'):
				track.addPuid(node.getAttribute('id'))

	def _addISRCs(self, isrcListNode, track):
		for node in _getChildElements(isrcListNode):
			if _matches(node, 'isrc') and node.hasAttribute('id'):
				track.addISRC(node.getAttribute('id'))

	def _addRelationsToEntity(self, relationListNode, entity):
		targetType = _getUriAttr(relationListNode, 'target-type', NS_REL_1)

		if targetType is None:
			return

		for node in _getChildElements(relationListNode):
			if _matches(node, 'relation'):
				rel = self._createRelation(node, targetType)
				if rel is not None:
					entity.addRelation(rel)


	def _createRelation(self, relationNode, targetType):
		relation = self._factory.newRelation()

		relation.setType(_getUriAttr(relationNode, 'type', NS_REL_1))
		relation.setTargetType(targetType)
		resType = _getResourceType(targetType)
		relation.setTargetId(_getIdAttr(relationNode, 'target', resType))

		if relation.getType() is None \
				or relation.getTargetType() is None \
				or relation.getTargetId() is None:
			return None

		relation.setDirection(_getDirectionAttr(relationNode, 'direction'))
		relation.setBeginDate(_getDateAttr(relationNode, 'begin'))
		relation.setEndDate(_getDateAttr(relationNode, 'end'))

		for a in _getUriListAttr(relationNode, 'attributes', NS_REL_1):
			relation.addAttribute(a)

		target = None
		children = _getChildElements(relationNode)
		if len(children) > 0:
			node = children[0]
			if _matches(node, 'artist'):
				target = self._createArtist(node)	
			elif _matches(node, 'release'):
				target = self._createRelease(node)	
			elif _matches(node, 'track'):
				target = self._createTrack(node)	

		relation.setTarget(target)

		return relation


#
# XML output
#

class _XmlWriter(object):
	def __init__(self, outStream, indentAmount='  ', newline="\n"):
		self._out = outStream
		self._indentAmount = indentAmount
		self._stack = [ ]
		self._newline = newline

	def prolog(self, encoding='UTF-8', version='1.0'):
		pi = '<?xml version="%s" encoding="%s"?>' % (version, encoding)
		self._out.write(pi + self._newline)

	def start(self, name, attrs={ }):
		indent = self._getIndention()
		self._stack.append(name)
		self._out.write(indent + self._makeTag(name, attrs) + self._newline)

	def end(self):
		name = self._stack.pop()
		indent = self._getIndention()
		self._out.write('%s</%s>\n' % (indent, name))

	def elem(self, name, value, attrs={ }):
		# delete attributes with an unset value
		for (k, v) in attrs.items():
			if v is None or v == '':
				del attrs[k]

		if value is None or value == '':
			if len(attrs) == 0:
				return
			self._out.write(self._getIndention())
			self._out.write(self._makeTag(name, attrs, True) + '\n')
		else:
			escValue = saxutils.escape(value or '')
			self._out.write(self._getIndention())
			self._out.write(self._makeTag(name, attrs))
			self._out.write(escValue)
			self._out.write('</%s>\n' % name)

	def _getIndention(self):
		return self._indentAmount * len(self._stack)

	def _makeTag(self, name, attrs={ }, close=False):
		ret = '<' + name

		for (k, v) in attrs.iteritems():
			if v is not None:
				v = saxutils.quoteattr(str(v))
				ret += ' %s=%s' % (k, v)

		if close:
			return ret + '/>'
		else:
			return ret + '>'



class MbXmlWriter(object):
	"""Write XML in the Music Metadata XML format."""

	def __init__(self, indentAmount='  ', newline="\n"):
		"""Constructor.

		@param indentAmount: the amount of whitespace to use per level
		"""
		self._indentAmount = indentAmount
		self._newline = newline


	def write(self, outStream, metadata):
		"""Writes the XML representation of a Metadata object to a file.

		@param outStream: an open file-like object
		@param metadata: a L{Metadata} object
		"""
		xml = _XmlWriter(outStream, self._indentAmount, self._newline)

		xml.prolog()
		xml.start('metadata', {
			'xmlns': NS_MMD_1,
			'xmlns:ext': NS_EXT_1,
		})

		self._writeArtist(xml, metadata.getArtist())
		self._writeRelease(xml, metadata.getRelease())
		self._writeReleaseGroup(xml, metadata.getReleaseGroup())
		self._writeTrack(xml, metadata.getTrack())
		self._writeLabel(xml, metadata.getLabel())

		if len(metadata.getArtistResults()) > 0:
			xml.start('artist-list', {
				'offset': metadata.artistResultsOffset,
				'count': metadata.artistResultsCount,
			})
			for result in metadata.getArtistResults():
				self._writeArtist(xml, result.getArtist(),
					result.getScore())
			xml.end()

		if len(metadata.getReleaseResults()) > 0:
			xml.start('release-list', {
				'offset': metadata.releaseResultsOffset,
				'count': metadata.releaseResultsCount,
			})
			for result in metadata.getReleaseResults():
				self._writeRelease(xml, result.getRelease(),
					result.getScore())
			xml.end()

		if len(metadata.getReleaseGroupResults()) > 0:
			xml.start('release-group-list', {
				'offset': metadata.releaseGroupResultsOffset,
				'count': metadata.releaseGroupResultsCount
			})
			for result in metadata.getReleaseGroupResults():
				self._writeReleaseGroup(xml, result.getReleaseGroup(),
					result.getScore())
			xml.end()

		if len(metadata.getTrackResults()) > 0:
			xml.start('track-list', {
				'offset': metadata.trackResultsOffset,
				'count': metadata.trackResultsCount,
			})
			for result in metadata.getTrackResults():
				self._writeTrack(xml, result.getTrack(),
					result.getScore())
			xml.end()

		if len(metadata.getLabelResults()) > 0:
			xml.start('label-list', {
				'offset': metadata.labelResultsOffset,
				'count': metadata.labelResultsCount,
			})
			for result in metadata.getLabelResults():
				self._writeLabel(xml, result.getLabel(),
					result.getScore())
			xml.end()

		xml.end()


	def _writeArtist(self, xml, artist, score=None):
		if artist is None:
			return

		xml.start('artist', {
			'id': mbutils.extractUuid(artist.getId()),
			'type': mbutils.extractFragment(artist.getType()),
			'ext:score': score,
		})

		xml.elem('name', artist.getName())
		xml.elem('sort-name', artist.getSortName())
		xml.elem('disambiguation', artist.getDisambiguation())
		xml.elem('life-span', None, {
			'begin': artist.getBeginDate(),
			'end': artist.getEndDate(),
		})

		if len(artist.getAliases()) > 0:
			xml.start('alias-list')
			for alias in artist.getAliases():
				xml.elem('alias', alias.getValue(), {
					'type': alias.getType(),
					'script': alias.getScript(),
				})
			xml.end()

		if len(artist.getReleases()) > 0:
			xml.start('release-list')
			for release in artist.getReleases():
				self._writeRelease(xml, release)
			xml.end()

		if len(artist.getReleaseGroups()) > 0:
			xml.start('release-group-list')
			for releaseGroup in artist.getReleaseGroups():
				self._writeReleaseGroup(xml, releaseGroup)
			xml.end()

		self._writeRelationList(xml, artist)
		# TODO: extensions

		xml.end()


	def _writeRelease(self, xml, release, score=None):
		if release is None:
			return

		types = [mbutils.extractFragment(t) for t in release.getTypes()]
		typesStr = None
		if len(types) > 0:
			typesStr = ' '.join(types)

		xml.start('release', {
			'id': mbutils.extractUuid(release.getId()),
			'type': typesStr,
			'ext:score': score,
		})

		xml.elem('title', release.getTitle())
		xml.elem('text-representation', None, {
			'language': release.getTextLanguage(),
			'script': release.getTextScript()
		})
		xml.elem('asin', release.getAsin())

		self._writeArtist(xml, release.getArtist())
		self._writeReleaseGroup(xml, release.getReleaseGroup())

		if len(release.getReleaseEvents()) > 0:
			xml.start('release-event-list')
			for event in release.getReleaseEvents():
				self._writeReleaseEvent(xml, event)
			xml.end()

		if len(release.getDiscs()) > 0:
			xml.start('disc-list')
			for disc in release.getDiscs():
				xml.elem('disc', None, { 'id': disc.getId() })
			xml.end()

		if len(release.getTracks()) > 0:
			# TODO: count attribute
			xml.start('track-list', {
				'offset': release.getTracksOffset()
			})
			for track in release.getTracks():
				self._writeTrack(xml, track)
			xml.end()
		
		self._writeRelationList(xml, release)
		# TODO: extensions

		xml.end()

	def _writeReleaseGroup(self, xml, rg, score = None):
		if rg is None:
			return

		xml.start('release-group', {
			'id': mbutils.extractUuid(rg.getId()),
			'type': mbutils.extractFragment(rg.getType()),
			'ext:score': score,
		})

		xml.elem('title', rg.getTitle())
		self._writeArtist(xml, rg.getArtist())

		if len(rg.getReleases()) > 0:
			xml.start('release-list')
			for rel in rg.getReleases():
				self._writeRelease(xml, rel)
			xml.end()

		xml.end()

	def _writeReleaseEvent(self, xml, event):
		xml.start('event', {
			'country': event.getCountry(),
			'date': event.getDate(),
			'catalog-number': event.getCatalogNumber(),
			'barcode': event.getBarcode(),
			'format': event.getFormat()
		})

		self._writeLabel(xml, event.getLabel())

		xml.end()


	def _writeTrack(self, xml, track, score=None):
		if track is None:
			return

		xml.start('track', {
			'id': mbutils.extractUuid(track.getId()),
			'ext:score': score,
		})
		
		xml.elem('title', track.getTitle())
		xml.elem('duration', str(track.getDuration()))
		self._writeArtist(xml, track.getArtist())

		if len(track.getReleases()) > 0:
			# TODO: offset + count
			xml.start('release-list')
			for release in track.getReleases():
				self._writeRelease(xml, release)
			xml.end()

		if len(track.getPuids()) > 0:
			xml.start('puid-list')
			for puid in track.getPuids():
				xml.elem('puid', None, { 'id': puid })
			xml.end()

		self._writeRelationList(xml, track)
		# TODO: extensions

		xml.end()


	def _writeLabel(self, xml, label, score=None):
		if label is None:
			return

		xml.start('label', {
			'id': mbutils.extractUuid(label.getId()),
			'type': mbutils.extractFragment(label.getType()),
			'ext:score': score,
		})

		xml.elem('name', label.getName())
		xml.elem('sort-name', label.getSortName())
		xml.elem('disambiguation', label.getDisambiguation())
		xml.elem('life-span', None, {
			'begin': label.getBeginDate(),
			'end': label.getEndDate(),
		})

		if len(label.getAliases()) > 0:
			xml.start('alias-list')
			for alias in label.getAliases():
				xml.elem('alias', alias.getValue(), {
					'type': alias.getType(),
					'script': alias.getScript(),
				})
			xml.end()

		# TODO: releases, artists

		self._writeRelationList(xml, label)
		# TODO: extensions

		xml.end()


	def _writeRelationList(self, xml, entity):
		for tt in entity.getRelationTargetTypes():
			xml.start('relation-list', {
				'target-type': mbutils.extractFragment(tt),
			})
			for rel in entity.getRelations(targetType=tt):
				self._writeRelation(xml, rel, tt)
			xml.end()


	def _writeRelation(self, xml, rel, targetType):
		relAttrs = ' '.join([mbutils.extractFragment(a) 
				for a in rel.getAttributes()])

		if relAttrs == '':
			relAttrs = None

		attrs = {
			'type': mbutils.extractFragment(rel.getType()),
			'target': rel.getTargetId(),
			'direction': rel.getDirection(),
			'begin': rel.getBeginDate(),
			'end': rel.getBeginDate(),
			'attributes': relAttrs,
		}

		if rel.getTarget() is None:
			xml.elem('relation', None, attrs)
		else:
			xml.start('relation', attrs)
			if targetType == NS_REL_1 + 'Artist':
				self._writeArtist(xml, rel.getTarget())
			elif targetType == NS_REL_1 + 'Release':
				self._writeRelease(xml, rel.getTarget())
			elif targetType == NS_REL_1 + 'Track':
				self._writeTrack(xml, rel.getTarget())
			xml.end()
			

#
# DOM Utilities
#

def _matches(node, name, namespace=NS_MMD_1):
	"""Checks if an xml.dom.Node and a given name and namespace match."""

	if node.localName == name and node.namespaceURI == namespace:
		return True
	else:
		return False


def _getChildElements(parentNode):
	"""Returns all direct child elements of the given xml.dom.Node."""

	children = [ ]
	for node in parentNode.childNodes:
		if node.nodeType == node.ELEMENT_NODE:
			children.append(node)

	return children


def _getText(element, regex=None, default=None):
	"""Returns the text content of the given xml.dom.Element.

	This function simply fetches all contained text nodes, so the element
	should not contain child elements.
	"""
	res = ''
	for node in element.childNodes:
		if node.nodeType == node.TEXT_NODE:
			res += node.data

	if regex is None or re.match(regex, res):
		return res
	else:
		return default


def _getPositiveIntText(element):
	"""Returns the text content of the given xml.dom.Element as an int."""

	res = _getText(element)

	if res is None:
		return None

	try:
		return int(res)
	except ValueError:
		return None


def _getAttr(element, attrName, regex=None, default=None, ns=None):
	"""Returns an attribute of the given element.

	If there is no attribute with that name or the attribute doesn't
	match the regular expression, default is returned.
	"""
	if element.hasAttributeNS(ns, attrName):
		content = element.getAttributeNS(ns, attrName)

		if regex is None or re.match(regex, content):
			return content
		else:
			return default
	else:
		return default


def _getDateAttr(element, attrName):
	"""Gets an incomplete date from an attribute."""
	return _getAttr(element, attrName, '^\d+(-\d\d)?(-\d\d)?$')


def _getIdAttr(element, attrName, typeName):
	"""Gets an ID from an attribute and turns it into an absolute URI."""
	value = _getAttr(element, attrName)

	return _makeAbsoluteUri('http://musicbrainz.org/' + typeName + '/', value)

	

def _getIntAttr(element, attrName, min=0, max=None, ns=None):
	"""Gets an int from an attribute, or None."""
	try:
		val = int(_getAttr(element, attrName, ns=ns))

		if max is None:
			max = val

		if min <= val <= max:
			return val
		else:
			return None
	except ValueError:
		return None # raised if conversion to int fails
	except TypeError:
		return None # raised if no such attribute exists


def _getUriListAttr(element, attrName, prefix=NS_MMD_1):
	"""Gets a list of URIs from an attribute."""
	if not element.hasAttribute(attrName):
		return [ ]

	f = lambda x: x != ''
	uris = filter(f, re.split('\s+', element.getAttribute(attrName)))

	m = lambda x: _makeAbsoluteUri(prefix, x)
	uris = map(m, uris)

	return uris


def _getUriAttr(element, attrName, prefix=NS_MMD_1):
	"""Gets a URI from an attribute.

	This also works for space-separated URI lists. In this case, the
	first URI is returned.
	"""
	uris = _getUriListAttr(element, attrName, prefix)
	if len(uris) > 0:
		return uris[0]
	else:
		return None


def _getBooleanAttr(element, attrName):
	"""Gets a boolean value from an attribute."""
	value = _getAttr(element, attrName)
	if value == 'true':
		return True
	elif value == 'false':	
		return False
	else:
		return None


def _getDirectionAttr(element, attrName):
	"""Gets the Relation reading direction from an attribute."""
	regex = '^\s*(' + '|'.join((
				model.Relation.DIR_FORWARD,
				model.Relation.DIR_BACKWARD)) + ')\s*$'
	return _getAttr(element, 'direction', regex, model.Relation.DIR_NONE)


def _makeAbsoluteUri(prefix, uriStr):
	"""Creates an absolute URI adding prefix, if necessary."""
	if uriStr is None:
		return None

	(scheme, netloc, path, params, query, frag) = urlparse.urlparse(uriStr)

	if scheme == '' and netloc == '':
		return prefix + uriStr
	else:
		return uriStr
 
 
def _getResourceType(uri):
	"""Gets the resource type from a URI.

	The resource type is the basename of the URI's path.
	"""
	m = re.match('^' + NS_REL_1 + '(.*)$', uri)
	
	if m:
		return m.group(1).lower()
	else:
		return None

# EOF



"""Tests for the MbXmlParser class."""
import unittest
from musicbrainz2.wsxml import MbXmlParser, ParseError, _makeAbsoluteUri
import StringIO
import os.path

VALID_DATA_DIR = os.path.join('test-data', 'valid')
INVALID_DATA_DIR = os.path.join('test-data', 'invalid')

VALID_ARTIST_DIR = os.path.join(VALID_DATA_DIR, 'artist')
INVALID_ARTIST_DIR = os.path.join(INVALID_DATA_DIR, 'artist')


class BaseParserTest(unittest.TestCase):
	"""Test the most basic cases: empty and invalid documents."""
	
	def __init__(self, name):
		unittest.TestCase.__init__(self, name)


	def testEmptyValid(self):
		tests = ('empty_1.xml', 'empty_2.xml')

		for f in self._makeFiles(VALID_ARTIST_DIR, tests):
			try:
				MbXmlParser().parse(f)
			except ParseError, e:
				self.fail(f.name + ' ' + e.msg)


	def testEmptyInvalid(self):
		tests = ('empty_1.xml', 'empty_2.xml', 'empty_3.xml')

		for f in self._makeFiles(INVALID_ARTIST_DIR, tests):
			p = MbXmlParser()
			try:
				MbXmlParser().parse(f)
			except ParseError:
				pass
			else:
				self.fail(f.name + ' ' + f.name)


	def testMakeAbsoluteUri(self):
		self.assert_(_makeAbsoluteUri('http://mb.org/artist/', None) is None)
		self.assertEquals('http://mb.org/artist/some_id',
			_makeAbsoluteUri('http://mb.org/artist/', 'some_id'))
		self.assertEquals('http://mb.org/artist/some_id',
			_makeAbsoluteUri('http://mb.org/artist/',
				'http://mb.org/artist/some_id'))
		self.assertEquals('http://mb.org/ns/mmd-1.0#name',
			_makeAbsoluteUri('http://mb.org/ns/mmd-1.0#', 'name'))
		self.assertEquals('http://mb.org/ns/mmd-1.0#name',
			_makeAbsoluteUri('http://mb.org/ns/mmd-1.0#',
				'http://mb.org/ns/mmd-1.0#name'))


	def _makeFiles(self, dir, basenames):
		files = [ ]
		for b in basenames:
			files.append( file(os.path.join(dir, b)) )

		return files


# EOF
